import matplotlib.pyplot as plt
import pickle
import time
import numpy as np
import useful as uf
from . import nanonis
import ast
from matplotlib.colors import LinearSegmentedColormap
import matplotlib

from scipy.ndimage import gaussian_filter
from scipy.interpolate import interp1d
spectra = nanonis.biasSpectroscopy()

def findmax(fnames):
    energy = []
    plt.figure()
    map = []
    for f in fnames:
        y = spectra.load(f)
        x, y = uf.data_smooth(spectra.bias,spectra.conductance,order=10)
        energy.append(x[y.argmax()])
        map.append(y)
    plt.imshow(map,extent=[spectra.bias.min(),spectra.bias.max(),0,1],aspect='auto',interpolation='nearest')
    return np.array(energy)


def dbm_mv(x):
    return 10**((x+47)/20) 


def mv_dbm(x):
    return -47+20*np.log10(x)


def TF_fast(fnames1,fnames2,p1,p2,TF_energy):
    ''' Outputs the correct power given two sets of spectras at 2 diferent powers
    
    Inputs are 2 lists of paths to the files .dat, 2 power arrays, and wanted energy of the outter coherence peak
    ''' 
    x0s = np.abs(sp.findmax(fnames1))
    x1s = np.abs(sp.findmax(fnames2))
    def line(x,m,q):
        return m*x+q

    def m_calc(x0,x1,p1,p2):
        return (p2-p1)/(x1-x0)

    def q_calc(p1,m,x0):
        return p1-m*x0

    new_P = []

    for i in range(10):
        m = m_calc(x0s[i],x1s[i],p1[i],p2[i])
        q = q_calc(p1[i],m,x0s[i])
        new_P.append(line(TF_energy,m,q))
    return new_P


import numpy as np
from scipy import constants as const
## PoE Superconductivity

Eh = const.physical_constants['Hartree energy in eV']

def f(t,x,R,C,T):
    x=np.linspace(0.01,5,2000)
    dx=x[1]-x[0]
    ReZ=R/(1+(x*R*C)**2)
    a,b=np.meshgrid(x,t)
    c=a*b
    return  np.dot(np.cos(c)-1,2*(ReZ/x)/np.tanh(x/(2*T)))*dx- np.dot(1j*np.sin(c) ,2*(ReZ/x) )*dx


def PD(R,C1,C2,T,n,N,Vmax):
    dis=(N/2-1)/(16*np.pi*Vmax*Eh)

    C=C1*C2/(C1+C2)
    if n==1:
        k=C2/(C1+C2)
    else:
        k=C1/(C1+C2)
    t=np.linspace(-dis,dis,N)
    x=np.linspace(0.01,10,100)
    freq=np.fft.fftfreq(N,d=2*dis/N)
    a=np.argsort(freq)
    y=np.real(np.fft.fft(np.exp(k*f(t,x,R,C,T))))
    t=np.abs(y[a])
    return np.flip(np.sort(freq)),t/sum(t)

def initD_(R1,R2,C1,C2,V,delta_t,Delta,T,eta,N):
    global EnD
    EnD=[0 for i in range(25)]
    global MnD
    MnD=[0 for i in range(25)]
    global x_int
    x_int=np.linspace(-4*np.max(V),4*np.max(V),N)
    a,b=np.meshgrid(x_int,x_int)
    t=a+b
    global bcsf
    bcsf=bcs(delta_t,t,eta)*fermi(T,t)

    global bcsif_even
    bcsif_even=bcs_i(Delta,t,eta,0)*fermi(T,t)

    global bcsif_odd
    bcsif_odd=bcs_i(Delta,t,eta,1)*fermi(T,t)

    f1,a1=PD(R1+R2,C1,C2,T*Eh,1,N,np.max(V))
    global freq1
    freq1=f1/(2*np.pi*27)
    global PoE1
    PoE1=a1
    f2,a2=PD(R1+R2,C1,C2,T*Eh,2,N,np.max(V))
    global freq2
    freq2=f2/(2*np.pi*27)
    global PoE2
    PoE2=a2

def bcs(delta,x,eta):
    if delta<=0.01: 
        return 1
    else:
        return (np.sign(x))*np.imag(np.divide(np.abs(x+eta*1j),np.sqrt(delta**2-(x+eta*1j)**2)))
    
def bcs_i(delta,x,eta,n):
    if np.mod(n,2)==0:
        if delta<=0.01: 
            return 1
        else:
            return (np.sign(x))*np.imag(np.divide(np.abs(x+eta*1j),np.sqrt(delta**2-(x+eta*1j)**2)))
    else:
        return 1
        

def fermi(T,x):
    if T==0.0:
        return np.heaviside(-x,1)
    else:
        return np.divide(1,1+np.exp(x/T))

def E1(C1,C2,V,n,Q0):
    k=C2/(C1+C2)
    return k*V+(n+Q0-1/2)/(C1+C2)

def E2(C1,C2,V,n,Q0):
    k=C1/(C1+C2)
    return k*V+(n+Q0-1/2)/(C1+C2)

def Gamma1D(V,R1,C1,C2,n,Q0,Delta,delta_t,delta_s,T,eta):
    a,b=np.meshgrid(x_int,E1(C1,C2,V,n,Q0))
    t=a+b
    return np.dot(bcs_i(Delta,t,eta,n)*(fermi(T,-t)),np.dot(bcsf,PoE1))/R1


def Gamma2D(V,R2,C1,C2,n,Q0,Delta,delta_t,delta_s,T,eta):
    a,b=np.meshgrid(x_int,E2(C1,C2,V,-n,-Q0))
    t=a+b
    if np.mod(n,2)==0:
        return np.dot(bcs_i(Delta,t,eta,n)*(fermi(T,-t)),np.dot(bcsif_even,PoE1))/R2
    else:
        return np.dot(bcs_i(Delta,t,eta,n)*(fermi(T,-t)),np.dot(bcsif_odd,PoE1))/R2


def PND(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta,n):
    mn=(Gamma1D(V,R1,C1,C2,-n,Q0,Delta,delta_t,delta_s,T,eta)+Gamma2D(-V,R2,C1,C2,n,-Q0,Delta,delta_t,delta_s,T,eta))/(Gamma1D(-V,R1,C1,C2,n+1,-Q0,Delta,delta_t,delta_s,T,eta)+Gamma2D(V,R2,C1,C2,-n-1,Q0,Delta,delta_t,delta_s,T,eta))
    en=(Gamma1D(-V,R1,C1,C2,-n,-Q0,Delta,delta_t,delta_s,T,eta)+Gamma2D(V,R2,C1,C2,n,Q0,Delta,delta_t,delta_s,T,eta))/(Gamma1D(V,R1,C1,C2,n+1,Q0,Delta,delta_t,delta_s,T,eta)+Gamma2D(-V,R2,C1,C2,-n-1,-Q0,Delta,delta_t,delta_s,T,eta))
    global En
    EnD[n]= en
    global Mn
    MnD[n]= mn
    return en,mn

def check_pD(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta):
    n=[0]
    while True:
        a=[]
        b=[]
        for i in n:
            an,bn=PND(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta,i)
            a.append(an[0])
            b.append(bn[0])
        p0=1
        for i in range(len(a)):
            temp=np.zeros(len(a))
            temp[0:i+1]=1
            temp=temp.tolist()
            p0+=np.prod(a,where=temp)+np.prod(b,where=temp)
        p0=1/p0
        pn=p0*np.prod(a)
        p_n=p0*np.prod(b)
        #print(pn,p_n)
        n.append(n[-1]+1)
        if pn<0.01 and p_n<0.01:
            #print(len(a))
            break 
    return n[-1]

def all_pD(V,n):
    a=[]
    b=[]
    for i in range(n+1):
        a.append(EnD[i])
        b.append(MnD[i])
    p0=1
    pn=[]
    p_n=[]
    for i in range(n):
        temp1=np.full(len(V),1.0)
        temp2=np.full(len(V),1.0)
        for j in range(i+1):
            temp1*=a[j]     
            temp2*=b[j]
        p0+=temp1+temp2
        pn.append(temp1)
        p_n.append(temp2)
    p0=1/p0
    pn=p0*np.array(pn)
    p_n=p0*np.array(p_n)
    return p0,pn,p_n

def G1nD(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta,n):
    return Gamma1D(V,R1,C1,C2,n,Q0,Delta,delta_t,delta_s,T,eta)-Gamma1D(-V,R1,C1,C2,-n,-Q0,Delta,delta_t,delta_s,T,eta)

def G2nD(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta,n):
    return Gamma2D(V,R2,C1,C2,n,Q0,Delta,delta_t,delta_s,T,eta)-Gamma2D(-V,R2,C1,C2,-n,-Q0,Delta,delta_t,delta_s,T,eta)

def currentD(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta):
    initD_(R1,R2,C1,C2,V,delta_t,Delta,T,eta,5000)
    n=check_pD(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta)
    p0,pn,p_n=all_pD(V,n)
    I=p0*G1nD(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta,0)
    for i in range(1,n+1):
        I+=pn[i-1]*G1nD(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta,i)
        I+=p_n[i-1]*G1nD(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta,-i)
    t=np.gradient(I)
    return t/np.sum(t)


def currentD2(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta):
    initD_(R1,R2,C1,C2,V,delta_t,Delta,T,eta,5000)
    n=check_pD(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta)
    p0,pn,p_n=all_pD(V,n)
    I=p0*G2nD(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta,0)
    for i in range(1,n+1):
        I+=pn[i-1]*G2nD(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta,i)
        I+=p_n[i-1]*G2nD(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta,-i)
    t=np.gradient(I)
    return t/np.sum(t)




import matplotlib as plt
import os
import csv
import numpy as np
from PyQt5 import QtCore, QtWidgets
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure, Artist

plt.rcParams['lines.linewidth'] = 1.0
plt.rcParams['font.family'] = 'sans-serif'
plt.rcParams['font.sans-serif'] = 'Arial'
plt.rcParams['font.size'] = 10
plt.use('Qt5Agg')

def saveCSV(array1,array2,filename):
    matrix = np.vstack((array1,array2))
    matrix = np.transpose(matrix)
    filename_mod = filename[:-7] + 'Dec' + filename[-7:]
    with open(filename_mod, 'w') as csvfile:
        writer = csv.writer(csvfile)
        [writer.writerow(r) for r in matrix]


class Canvas(FigureCanvas):

    def __init__(self, parent=None, width=7, height =5, dpi=100, tight_layout = False):
        self.fig = Figure(figsize=(width, height), dpi=dpi)
        self.fig.subplots_adjust(hspace=.5)
        self.ax1 = self.fig.add_subplot(211)
        self.ax2 = self.fig.add_subplot(212)

        FigureCanvas.__init__(self,self.fig)
        self.setParent(None)

        FigureCanvas.setSizePolicy(self,
                                   QtWidgets.QSizePolicy.Expanding,
                                   QtWidgets.QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)



class MplCanvas(FigureCanvas):

    def __init__(self, parent=None, width=7, height =5, dpi=100, tight_layout = False):
        self.fig = Figure(figsize=(width, height), dpi=dpi)
        self.fig.subplots_adjust(hspace=.5)
        self.ax1 = self.fig.add_subplot(211)
        self.ax2 = self.fig.add_subplot(212)
        x = np.linspace(1,2,10)
        self.ax2.plot(x,x)

        FigureCanvas.__init__(self,self.fig)
        self.setParent(parent)

        FigureCanvas.setSizePolicy(self,
                                   QtWidgets.QSizePolicy.Expanding,
                                   QtWidgets.QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)

    def plot_filename(self, filename):
        route,file = os.path.split(filename)
        self.fig.suptitle(file)

    def plot_top(self, xAxis, yAxis, xLabel, yLabel, ref):
        self.ax1.plot(xAxis, yAxis)
        self.ax1.axis('on')
        self.ax1.grid('on')
        self.ax1.set_xlim(left=min(ref),right = max(ref))
        self.ax1.set_xlabel(xLabel)
        self.ax1.set_ylabel(yLabel)
        self.draw()

    def plot_bottom(self, xAxis, yAxis, xLabel, yLabel, ref):
        self.ax2.plot(xAxis, yAxis)
        self.ax2.axis('on')
        self.ax2.grid('on')
        self.ax2.set_xlim(left=min(ref),right = max(ref))
        self.ax2.set_xlabel(xLabel)
        self.ax2.set_ylabel(yLabel)
        self.draw()

    def clear_top(self):
        self.ax1.clear()
        self.draw()

    def clear_bottom(self):
        self.ax2.clear()
        self.draw()

class DSButton(QtWidgets.QLabel):

    def __init__(self, name, layout, parent, default=0, deci=4, maximum = 100, minimum = 0):
        
        self.label = QtWidgets.QLabel(name, parent)
        self.input = QtWidgets.QDoubleSpinBox(parent, decimals = deci)
        self.input.setRange(minimum, maximum)
        self.input.setSingleStep(1E-4)
        self.input.setValue(default)
        layout.addWidget(self.label)
        layout.addWidget(self.input)

    def get_value(self, factor):
        return self.input.value()/factor

class Tab(QtWidgets.QWidget):

    def __init__(self, name, parent):
        super(Tab, self).__init__()
        parent.addTab(self,name)
        self.layout = QtWidgets.QVBoxLayout()
        self.layout.setSpacing(10)
        self.layout.addStretch(0)
        self.setLayout(self.layout)


import detect_peaks as dp
from . import nanonis
#from modules import functions
import distributions
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import tkinter as tk
from tkinter import filedialog
from matplotlib import rc
from scipy.signal import savgol_filter
import glob
import colorcet as cc
import csv
import numpy as np
import pickle
from lmfit.models import LorentzianModel, ConstantModel, GaussianModel, PolynomialModel
plt.rcParams.update({'font.size': 12})
plt.rcParams["figure.figsize"] = [10,5]
#rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
rc('font',**{'family':'serif','serif':['Helvetica']})
# plt.switch_backend('Qt5Agg')
colour_code =cc.glasbey_category10


########################################
#README
#This package allows multi fit of several spectras.
#You can use it if you want to follow a peak shifting across different spectra.
# How to:
#1) gather filenames in a filenames list
#2) convert them with specTodf(filenames) procedure
#3) use one of the fit_plot type of functions. Input all the parameters and the fit will be executed.
#4) the output is a list of fit output (lmfit style), you can extract than the parameters you need with one of the extract functions.
##########################################


spectra = nanonis.biasSpectroscopy()

def find_nearest_idx(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx
def cutSpec(x,y,center,interval):
    x = np.array(x)
    y = np.array(y)
    startidx = find_nearest_idx(x,center+interval)
    endidx = find_nearest_idx(x,center-interval)
    x = x[startidx:endidx]
    y = y[startidx:endidx]
    return x, y
def save_obj(obj, name ):
    with open(name + '.pkl', 'wb') as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL) 
def load_obj(name ):
    with open( name + '.pkl', 'rb') as f:
        return pickle.load(f)
#converts a series of spectra (from filenames) to a dataframe compatible with the fitting modules
def specTodf(filenames,offset=0,norm=[-3e-3,3e-3]):
    data = pd.DataFrame()
    count = 0
    for filename in filenames:
        spectra.load(filename)
        spectra.normalizeRange(norm)
        spectra.biasOffset(offset)
        if count == 0:
            data[0]=spectra.bias*1e3
            data[1]=spectra.conductance
            count += 2
            continue
        data[count]=spectra.conductance
        count+=1
    return data

def arrayTodf(bias,conductance):
    data = pd.DataFrame()
    count = 0
    data[0]=bias*1e3
    for i in range(0,conductance.shape[0]):
        data[i+1]=conductance[i,:]
    return data
###################################################################
################ FUNCTIONS DEFINITIONS #############################
####################################################################
def fitFunc1g_const(x,y,center,sigma): #1 gaussian with constant background
    model = GaussianModel() + ConstantModel(prefix='bkg_')
    params = model.make_params()
    params['center'].set(center,vary=True)
    params['amplitude'].set(0.1,min=0,vary=True)
    params['sigma'].set(sigma,vary=True)
    params['sigma'].set(vary=True,max=0.1)
    params['bkg_c'].set(0,vary=True)
    result=model.fit(y,params,x=x)
    return result

def fitFunc1g_para(x,y,center,sigma): #1 gaussian with parabolic background
    model = GaussianModel() + PolynomialModel(2,prefix='bkg_')
    params = model.make_params()
    params['center'].set(center,vary=True)
    params['amplitude'].set(0.1,vary=True)
    params['sigma'].set(sigma,vary=True)
    params['bkg_c0'].set(0.1,min=0,vary=True)
    params['bkg_c1'].set(0)
    params['bkg_c2'].set(0,vary=False)
    result=model.fit(y,params,x=x)
    return result

def fitFunc1g_lin(x,y,center,sigma): #1 gaussian with parabolic background
    model = GaussianModel() + PolynomialModel(2,prefix='bkg_')
    params = model.make_params()
    params['center'].set(center,vary=True)
    params['amplitude'].set(0.1,vary=True)
    params['sigma'].set(sigma,vary=True)
    params['bkg_c0'].set(0.1,min=0,vary=True)
    params['bkg_c1'].set(0,min=0,vary=True)
    params['bkg_c2'].set(0,vary=False)
    result=model.fit(y,params,x=x)
    return result

## double peak fit for shiba n1 particle and hole
def fitFunc2g_para(x,y,c_out,c_in,sigma): #2 gaussians with parabolic background
    model = GaussianModel(prefix='out_')+ GaussianModel(prefix='in_') + PolynomialModel(2)
    
    params = model.make_params()
    ## Initialize parameters peak1
    params['out_center'].set(vary=True)
    params['out_center'].set(c_out)
    params['out_amplitude'].set(3, min=0)
    params['out_sigma'].set(sigma)
    params['out_sigma'].set(vary=False)
    ## Initialize parameters peak2
    params['in_center'].set(c_in)
    params['in_amplitude'].set(3, min=0)
    params['in_sigma'].set(sigma)
    params['in_sigma'].set(vary=False)
    ## Initialize parameters constant baseline
    params['c0'].set(2)
    params['c1'].set(2)
    params['c2'].set(5,vary=False)
    result=model.fit(y,params,x=x)
    return result

def fitFunc2g_const(x,y,c_out,c_in,sigma): #2 gaussians with constant background
    model = GaussianModel(prefix='out_')+ GaussianModel(prefix='in_') + PolynomialModel(1)
    
    params = model.make_params()
    ## Initialize parameters peak1
    params['out_center'].set(c_out,vary=True)
    params['out_amplitude'].set(1, min=0,vary=True)
    params['out_sigma'].set(sigma,vary=True)
    params['out_sigma'].set(vary=True)
    ## Initialize parameters peak2
    params['in_center'].set(c_in,vary=True)
    params['in_amplitude'].set(1, min=0,vary=True)
    params['in_sigma'].set(sigma,vary=True)
    params['in_sigma'].set(vary=True)
    ## Initialize parameters constant baseline
    params['c0'].set(0,vary=True)
    params['c1'].set(0,vary=True)
    result=model.fit(y,params,x=x)
    return result

def fitFunc2l_const(x,y,c_out,c_in,sigma): #2 lorentzian with constant background
    model = LorentzianModel(prefix='out_')+ LorentzianModel(prefix='in_') + PolynomialModel(1)
    params = model.make_params()
    ## Initialize parameters peak1
    params['out_center'].set(c_out)
    params['out_amplitude'].set(3, min=0)
    params['out_sigma'].set(sigma[0])
    params['out_sigma'].set(vary=False)
    ## Initialize parameters peak2
    params['in_center'].set(c_in)
    params['in_amplitude'].set(3, min=0)
    params['in_sigma'].set(sigma[1])
    params['in_sigma'].set(vary=False)
    ## Initialize parameters constant baseline
    params['c0'].set(2)
    params['c1'].set(2)
    result=model.fit(y,params,x=x)
    return result
    

####################################################################################################
######################### FUNCTIONS TO PERFORM AND PLOT THE FIT ####################################
####################################################################################################
def fit_plotfull(data,energy, interval,sigma,spac,bkg,plotype='fullPlot',plotGauss=True):
    fig, ax = plt.subplots(1,figsize=[8,3])
    i = 0
    multifit = []
    ######### FITTING ###########
    for column in data.columns:
        if column == 0:
            continue
        x , y = cutSpec(data[0],data[column],energy,interval)
        if bkg == 'para':
            fit_res = fitFunc1g_para(x,y,energy,sigma)
        elif bkg == 'lin':
            fit_res = fitFunc1g_lin(x,y,energy,sigma)
        else:
            fit_res = fitFunc1g_const(x,y,energy,sigma)
        ######PLOT RESULTS ###########
        if plotype == 'fullPlot':
            ax.plot(data[0],data[column]+i*spac)
        elif plotype == 'cutPlot':
            ax.plot(x,y+i*spac) #plot the spectras (ranges)
        xx=np.linspace(fit_res.params['center'].value-interval,fit_res.params['center'].value +interval,1000)
        ax.plot(xx,fit_res.eval(x=xx)+i*spac,color='r')
        comps = fit_res.eval_components() #for plotting the unconvoluted gaussians
        if plotGauss == True:
            for name, comp in comps.items():
                if 'bkg_c' in name:
                    continue
                ax.plot(x,comp+i*spac,color='b', label=name)
        i += 1
        
        multifit.append(fit_res.params)
    return multifit


## single peak fit with a set of initial energies
def fit_plot_Einput(data,energy, interval,sigma,spac,para,plotype):
    fig, ax = plt.subplots(1,figsize=[2,8])
    temp = []
    temp_S = []
    i = 0

    ######### FITTING ###########
    for column in data.columns:
        if column == 0:
            continue
        x , y = cutSpec(data[0],data[column],energy[i],interval)
        if para == True:
            fit_res = fitFunc1g_para(x,y,energy[i],sigma)
        else:
            fit_res = fitFunc1g_const(x,y,energy[i],sigma)
        ######PLOT RESULTS ###########
        if plotype == 'fullPlot':
            ax.plot(data[0],data[column]+i*spac)
        elif plotype == 'cutPlot':
            ax.plot(x,y+i*spac) #plot the spectras (ranges)
        xx=np.linspace(fit_res.params['center'].value-interval,fit_res.params['center'].value +interval,1000)
        ax.plot(xx,fit_res.eval(x=xx)+i*spac,color='r')
        i += 1
        ########### SAVE DATA ##########
        if fit_res.params['center'].stderr == None:
            fit_res.params['center'].stderr = 0
        temp.append(fit_res.params['center'].value)
        temp_S.append(fit_res.params['center'].stderr)
    return temp, temp_S



### centralE is the initial energy of the peaks (or the middle between two peaks)
### interval energy range you want to fit centered in centralE


#fit with 2 gaussian or lorentzian with linear or parabolic background returning all parameters
def fit_plot2gfull(data,centralE,c_out,c_in, interval,sigma,spac,plotype='fullPlot',plotGauss=True,plotfit=True,background='gauss_constant'):
    fig, ax = plt.subplots(1,figsize=[10,5])
    i=0
    multifit = []
    for column in data.columns:
        if column == 0: #jump if the column is bias
            continue
        x , y = cutSpec(data[0],data[column],centralE,interval) #cut the spectra

        if background == 'gauss_parabolic': #decide the background to use
            fit_res = fitFunc2g_para(x,y,c_out,c_in,sigma) #do the fit
        elif background == 'gauss_constant':
            fit_res = fitFunc2g_const(x,y,c_out,c_in,sigma) #do the fit
        elif background == 'lorentz_constant':
            fit_res = fitFunc2l_const(x,y,c_out,c_in,sigma)
        else:
            break

        comps = fit_res.eval_components() #for plotting the unconvoluted gaussians

        ####plot things
        if plotype == 'fullPlot':
            ax.plot(data[0],data[column]+i*spac)
        elif plotype == 'cutPlot':
            ax.plot(x,y+i*spac) #plot the spectras (ranges)
        xx=np.linspace(centralE-interval,centralE+interval,1000) #create an interpolated bias coordinate to plot the fit
        if plotfit==True:
            ax.plot(xx,fit_res.eval(x=xx)+i*spac,color='g')  #plot the fits
           ## plot the separated gaussians  
        if plotGauss == True:
            for name, comp in comps.items():
                if 'c' in name:
                    continue
                if 'in' in name:
                    color='b'
                if 'out' in name:
                    color='r'
                ax.plot(x,comp+i*spac,color=color, label=name)
        i += 1
        ######## save results checking for nones values
        if fit_res.params['out_center'].stderr == None:
            fit_res.params['out_center'].stderr = 0
        if fit_res.params['in_center'].stderr == None:
            fit_res.params['in_center'].stderr = 0
        multifit.append(fit_res.params)
    return multifit



####################################################################################################
#################### USEFUL FUNCTIONS CORRELATED WITH THIS PACKAGE #################################
####################################################################################################
def plotshibafit(data,sets,offset): #data contains the values of the fits, sets is a list of column names in string format
    fig, ax = plt.subplots(1,figsize=[5,5])
    plt.rcParams.update({'font.size': 12})
    i=0
    for item in sets:
        if i!=0:
            offset=0 
        ax.errorbar(data[item]+offset,data['d'],data[item + '_S'],marker='o', markersize=8, capsize=5)
        i+=1
    ax.set_xlabel('Bias (mV)')
    ax.set_ylabel('Distance (nm)')
    plt.tight_layout()


#to extract peak data from packed fit objects


#to extract parameters from list of fit results
def fitParamExtract2g(fit_result,param,in_or_out='in'):
    param_V = []
    param_S = []
    for i in fit_result:
        param_V.append(i[in_or_out+'_'+param].value)
        param_S.append(i[in_or_out+'_'+param].stderr)
    param_V = np.array(param_V)
    param_S = np.array(param_S)
    return param_V, param_S

#to extract parameters from list of fit results
def fitParamExtract1g(fit_result,param):
    param_V = []
    param_S = []
    for i in fit_result:
        param_V.append(i[param].value)
        param_S.append(i[param].stderr)
    param_V = np.array(param_V)
    param_S = np.array(param_S)
    return param_V, param_S

def E_extract(raw_data, peak):
    energy = []
    energy_S = []
    count = 0
    for i in raw_data[peak]:
        energy.append(i['center'].value)
        energy_S.append(i['center'].stderr)
        count += 1
    return energy, energy_S

def E_extract_2g(raw_data,peak,in_or_out):
    energy = []
    energy_S = []
    count = 0
    for i in raw_data[peak]:
        energy.append(i[in_or_out + 'center'].value)
        energy_S.append(i[in_or_out + 'center'].stderr)
    return energy, energy_S




    ############# OLD CODE ####################
''' def fit_plot2g(data,centralE,c_out,c_in, interval,sigma,spac,plotype,plotGauss,plotfit,background):
    fig, ax = plt.subplots(1,figsize=[2,16])
    i=0
    multifit = []
    for column in data.columns:
        if column == 0: #jump if the column is bias
            continue
        x , y = cutSpec(data[0],data[column],centralE,interval) #cut the spectra

        if background == 'para': #decide the background to use
            fit_res = fitFunc2g_para(x,y,c_out,c_in,sigma) #do the fit
        elif background == 'constant':
            fit_res = fitFunc2g_const(x,y,c_out,c_in,sigma) #do the fit
        else:
            break

        comps = fit_res.eval_components() #for plotting the unconvoluted gaussians

        ####plot things
        if plotype == 'fullPlot':
            ax.plot(data[0],data[column]+i*spac)
        elif plotype == 'cutPlot':
            ax.plot(x,y+i*spac) #plot the spectras (ranges)
        xx=np.linspace(centralE-interval,centralE+interval,1000) #create an interpolated bias coordinate to plot the fit
        if plotfit==True:
            ax.plot(xx,fit_res.eval(x=xx)+i*spac,color='r')  #plot the fits
           ## plot the separated gaussians  
        if plotGauss == True:
            for name, comp in comps.items():
                if 'c' in name:
                    continue
                if 'in' in name:
                    color='b'
                if 'out' in name:
                    color='r'
                ax.plot(x,comp+i*spac,color=color, label=name)
        i += 1
        ######## save results checking for nones values
        if fit_res.params['out_center'].stderr == None:
            fit_res.params['out_center'].stderr = 0
        if fit_res.params['in_center'].stderr == None:
            fit_res.params['in_center'].stderr = 0
        multifit.append(fit_res.params)
    return multifit



def fit_plot(data,energy=0, interval=0,sigma=0,spac=0,bkg='lin',plotype='fullPlot',plotGauss=True):
    fig, ax = plt.subplots(1,figsize=[8,3])
    cent = []
    cent_S = []
    amp = []
    amp_S = []
    i = 0

    ######### FITTING ###########
    for column in data.columns:
        if column == 0:
            continue
        x , y = cutSpec(data[0],data[column],energy,interval)
        if bkg == 'para':
            fit_res = fitFunc1g_para(x,y,energy,sigma)
        elif bkg == 'lin':
            fit_res = fitFunc1g_lin(x,y,energy,sigma)
        else:
            fit_res = fitFunc1g_const(x,y,energy,sigma)
        ######PLOT RESULTS ###########
        if plotype == 'fullPlot':
            ax.plot(data[0],data[column]+i*spac)
        elif plotype == 'cutPlot':
            ax.plot(x,y+i*spac) #plot the spectras (ranges)
        xx=np.linspace(fit_res.params['center'].value-interval,fit_res.params['center'].value +interval,1000)
        ax.plot(xx,fit_res.eval(x=xx)+i*spac,color='r')
        comps = fit_res.eval_components() #for plotting the unconvoluted gaussians
        if plotGauss == True:
            for name, comp in comps.items():
                if 'bkg_c' in name:
                    continue
                ax.plot(x,comp+i*spac,color='b', label=name)
        i += 1
        ########### SAVE DATA ##########
        if fit_res.params['center'].stderr == None:
            fit_res.params['center'].stderr = 0
        cent.append(fit_res.params['center'].value)
        cent_S.append(fit_res.params['center'].stderr)
        amp.append(fit_res.params['amplitude'].value)
        amp_S.append(fit_res.params['amplitude'].stderr)        
    return cent, cent_S, amp, amp_S
'''


import numpy as np
import distributions
import nanonis
import scipy.signal as signal
from lmfit.models import VoigtModel
from scipy import constants as const
import matplotlib.pyplot as plt

def filterG(data, points=40, p=0.5, sigma=10):
	window = distributions.gaussian(np.linspace(-50,50,points), sigma)
	window = window/max(window)
	filtered = signal.fftconvolve(window, data)
	filtered = (np.average(data) / np.average(filtered)) * filtered
	filtered = np.roll(filtered,-int(points/2))
	return filtered[:-points]

def filterC(data, points=40, p=0.5, sigma=10):
	window = distributions.cauchy(np.linspace(-50,50,points), sigma)
	window = window/max(window)
	filtered = signal.fftconvolve(window, data)
	filtered = (np.average(data) / np.average(filtered)) * filtered
	filtered = np.roll(filtered,-int(points/2))
	return filtered[:-points]

def normalize(data, range=1.7E-3):
	indices = np.where(np.logical_or(data.bias>=range, data.bias<=-range))
	normal_state = np.mean(data.conductance[indices[0]])
	return data.conductance/normal_state

def getOffset(data):
	indices = np.where(np.logical_and(data.bias>=-7.5E-4, data.bias<=7.5E-4))
	start_ind = indices[0][0]
	end_ind = indices[0][-1]
	p = np.polyfit(data.bias[start_ind:end_ind],data.current[start_ind:end_ind],7)
	p2 = np.polyder(p,2)
	p1d1 = np.poly1d(p)
	p1d2 = np.poly1d(p2)
	superpoli = p1d1-p1d2
	return min(abs(superpoli.r))

def getOffset2(data):
	indices = np.where(np.logical_and(data.bias>=-7.5E-4, data.bias<=7.5E-4))
	start_ind = indices[0][0]
	end_ind = indices[0][-1]
	bias = np.array(data.bias[start_ind:end_ind])
	current = np.array(data.current[start_ind:end_ind])
	zero = np.argmin(abs(current))
	bias1 = np.array(bias[:zero])
	bias2 = np.array(bias[zero:])
	curr1 = np.array(current[:zero])
	curr2 = np.array(current[zero:])
	with plt.style.context('ggplot'):
		ax1=plt.subplot(211)
		ax1.plot(bias1, curr1,'*')
		ax1.plot(abs(bias2)+0.6E-4, abs(curr2),'*')
		ax2 = plt.subplot(212)
		ax2.plot(np.correlate(curr1,abs(curr2),'same'),'*')
	plt.show()


def fitVoigt(data, filtered, indices):
	gmodel = VoigtModel()
	center = []
	height = []
	for i in indices:
		index = (np.abs(data.bias-filtered.bias[i])).argmin()
		if data.bias[index]>=-1.4E-3 and data.bias[index]<=1.4E-3:
			indexO = (np.abs(data.bias-data.bias[index]-1.5E-4)).argmin()
			indexF = (np.abs(data.bias-data.bias[index]+1.5E-4)).argmin()
			xdata = data.bias[indexO:indexF]
			ydata = data.conductance[indexO:indexF]
			params = gmodel.make_params(amplitude=ydata.max(),
                            center=xdata.mean(),
                            sigma=xdata.std())
			result = gmodel.fit(ydata, params, x=xdata)
			parameters = result.params.valuesdict()
			center = np.append(center, parameters['center'])
			height = np.append(height,parameters['height'])
	if len(center) == 2:
		suma = {'PP1':center[0],'PP2':center[1],'HP1':height[0],'HP2':height[1],'SPEC':data.filename}
		return suma

def rotatePoint(point, angle):
	x = point[0]*np.cos(np.deg2rad(angle))-point[1]*np.sin(np.deg2rad(angle))
	y = point[0]*np.sin(np.deg2rad(angle))+point[1]*np.cos(np.deg2rad(angle))
	newPoint = [x,y]
	return  newPoint

def coherence(x,D,Delta,T):
	a = np.exp(-x/(np.sqrt(const.hbar*D/(Delta*const.e))))
	b = np.exp(-x/(np.sqrt(const.hbar*D/(2*np.pi*const.k*T))))
	return a*b


from re import T
from scipy.signal import savgol_filter
from scipy.interpolate import interp1d
from .superconductor import dynes_curve, dynes_curve_diff, dynes_wrong,coulomb,dynesdos, fdd,dynes_javi
from distributions import fermiDirac, fermiDirac_diff
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
def find_nearest(array,value):
    idx = (np.abs(array-value)).argmin()
    return idx

def sum_data(bias,conductance,x_min,x_max):    
    #a=data[data==x_min].index[0]
    #b=data[data==x_max].index[0]
    a=find_nearest(bias,x_min)
    b=find_nearest(bias,x_max)
    mean1=0.0
    mean2=0.0
    for i in range(0,b):
        mean1=mean1+conductance[i]
    for i in range(a,len(bias)):
        mean2=mean2+conductance[i]

    mean1=mean1/float(b)
    mean2=mean2/(float(len(bias)-a))

    return [mean1,mean2]

def extensions(bias,conductance,x_min,x_max,N):
    sp=bias[1]-bias[0]
    mean=sum_data(bias,conductance,x_min,x_max)
    a=[bias[0]-sp]
    b=[bias[len(bias)-1]+sp]
    c=[mean[0]]
    d=[mean[1]]
    for i in range(N):
        a.insert(0,a[0]-sp)
        b.append(b[len(b)-1]+sp)
        c.append(mean[0])
        d.append(mean[1])
    a=pd.Series(a)
    b=pd.Series(b)
    c=pd.Series(c)
    d=pd.Series(d)
    bias=pd.Series(bias)
    conductance=pd.Series(conductance)
    bias=a.append(bias)
    bias=bias.append(b)
    conductance=c.append(conductance)
    conductance=conductance.append(d)
    return [bias,conductance]

# def dynesDeconvolute(bias,conductance, gap=1.30e-3, temperature=1.248, dynesParameter=40e-6, energyR=6E-3, spacing=50e-6,x_min=-3.0E-3,x_max=3.0E-3,N=1000, window=15,order=3,n=3000):
#     bias_new=extensions(bias,conductance,x_min,x_max,N)[0]
#     conductance_n=extensions(bias,conductance,x_min,x_max,N)[1]
#     conductance_new=savgol_filter(conductance_n,window,order)
#     x=np.linspace(min(bias_new),max(bias_new),n)
#     f2 = interp1d(bias_new, conductance_new, kind='cubic')
#     plt.figure()
#     plt.plot(x,f2(x))
#     #dec_bias = np.arange(bias_new[np.round(len(bias_new)/2)]-energyR, bias_new[np.round(len(bias_new)/2)]+energyR, spacing)
#     dec_bias = np.arange(-energyR,energyR, spacing)
    
#     M_E, M_eV = np.meshgrid(dec_bias,x)
#     M_N = dynes_curve(M_E,gap,dynesParameter)
#     M_NV = dynes_curve(M_E+M_eV,gap,dynesParameter)
#     M_dNV = dynes_curve_diff(M_E+M_eV,gap,dynesParameter)

#     M_f = fermiDirac(M_E, temperature)
#     M_fV = fermiDirac(M_E+M_eV, temperature)
#     M_dfV = fermiDirac_diff(M_E+M_eV, temperature)
#     M_g = M_dNV*(M_f-M_fV) - M_NV*M_dfV
#     M_g_pinv = np.linalg.pinv(M_g)
#     dec_conductance = np.dot(M_g_pinv,f2(x))
#     np.savetxt('x.txt',dec_bias)
#     np.savetxt('y.txt',dec_conductance)
#     return dec_bias,dec_conductance

def dynesDeconvolute(bias,conductance, gap=1.4E-3, temperature=1.7, dynesParameter=4E-7, energyR=4E-3, spacing=53.5E-5,x_min=-4E-3,x_max=4.0E-3,N=100, window=19,order=15,n=1000):
    bias_new=extensions(bias,conductance,x_min,x_max,N)[0]
    conductance_n=extensions(bias,conductance,x_min,x_max,N)[1]
    conductance_new=savgol_filter(conductance_n,window,order)
    x=np.linspace(min(bias_new),max(bias_new),n)
    f2 = interp1d(bias_new, conductance_new, kind='cubic')
    #deconvolutionBias = np.arange(bias_new[np.round(len(bias_new)/2)]-energyR, bias_new[np.round(len(bias_new)/2)]+energyR, spacing)
    deconvolutionBias = np.arange(-energyR,energyR, spacing)
    
    M_E, M_eV = np.meshgrid(deconvolutionBias,x)
    M_N = dynes_curve(M_E,gap,dynesParameter)
    M_NV = dynes_curve(M_E+M_eV,gap,dynesParameter)
    M_dNV = dynes_curve_diff(M_E+M_eV,gap,dynesParameter)

    M_f = fermiDirac(M_E, temperature)
    M_fV = fermiDirac(M_E+M_eV, temperature)
    M_dfV = fermiDirac_diff(M_E+M_eV, temperature)

    M_g = M_dNV*(M_f-M_fV) - M_NV*M_dfV
    M_g_pinv = np.linalg.pinv(M_g)
    deconvolutionConductance = np.dot(M_g_pinv,f2(x))
    return deconvolutionBias,deconvolutionConductance



def dynesDeconvolute_nof(bias,conductance,gap=1.25e-3, temperature=1.7, dynesParameter=40E-6, energyR=6E-3, spacing = 30e-6,x_min=-2.0E-3,x_max=2.0E-3,N=100):
    bias_new=extensions(bias,conductance,x_min,x_max,N)[0]
    conductance_new=extensions(bias,conductance,x_min,x_max,N)[1]

    #deconvolutionBias = np.arange(bias_new[np.round(len(bias_new)/2)]-energyR, bias_new[np.round(len(bias_new)/2)]+energyR, spacing)
    dec_bias = np.arange(-energyR,energyR, spacing)
    
    M_E, M_eV = np.meshgrid(dec_bias,bias_new)
    #print(conductance_new)
    M_N = dynes_curve(M_E,gap,dynesParameter)
    M_NV = dynes_curve(M_E+M_eV,gap,dynesParameter)
    #M_dN = dynes_curve_diff(M_E,gap,dynesParameter)
    M_dNV = dynes_curve_diff(M_E+M_eV,gap,dynesParameter)

    M_f = fermiDirac(M_E, temperature)
    M_fV = fermiDirac(M_E+M_eV, temperature)
    #M_df = fermiDirac_diff(M_E, temperature)
    M_dfV = fermiDirac_diff(M_E+M_eV, temperature)
    M_g = M_dNV*(M_f-M_fV) - M_NV*M_dfV
    #g_conv = np.trapz((M_g*M_N),deconvolutionBias)
    M_g_pinv = np.linalg.pinv(M_g)
    dec_conductance = np.dot(M_g_pinv,conductance_new)
    #g_reconv = np.dot(M_g,deconvolutionConductance)
    return dec_bias,dec_conductance



def coulombDeconvolute_nof(bias,conductance,T,R2,C1,C2,Q0, energyR=6E-3, spacing = 30e-6,x_min=-2.0E-3,x_max=2.0E-3,N=100):
    bias_new=extensions(bias,conductance,x_min,x_max,N)[0]
    conductance_new=extensions(bias,conductance,x_min,x_max,N)[1]

    #deconvolutionBias = np.arange(bias_new[np.round(len(bias_new)/2)]-energyR, bias_new[np.round(len(bias_new)/2)]+energyR, spacing)
    dec_bias = np.arange(-energyR,energyR, spacing)
    
    M_E, M_eV = np.meshgrid(dec_bias,bias_new)
    #print(conductance_new)

    M_NV = coulomb(R2,C1,C2,Q0)
    #M_dN = dynes_curve_diff(M_E,gap,dynesParameter)
    M_dNV = coulomb(R2,C1,C2,Q0)

    M_f = fermiDirac(M_E, T)
    M_fV = fermiDirac(M_E+M_eV, T)
    #M_df = fermiDirac_diff(M_E, temperature)
    M_dfV = fermiDirac_diff(M_E+M_eV, T)
    M_g = M_dNV*(M_f-M_fV) - M_NV*M_dfV
    #g_conv = np.trapz((M_g*M_N),deconvolutionBias)
    M_g_pinv = np.linalg.pinv(M_g)
    dec_conductance = np.dot(M_g_pinv,conductance_new)
    #g_reconv = np.dot(M_g,deconvolutionConductance)
    return dec_bias,dec_conductance



def dynesConvolute(V,E_int,conductance,delta,T,gamma):
    curr = []
    for Vp in V:
        currp = np.trapz((conductance)*dynesdos(E_int-Vp,gamma,delta)*(fdd(E_int, Vp, T)-fdd(E_int,0, T)),x=E_int)
        curr.append(currp)
    return np.gradient(np.array(curr))





import matplotlib.pyplot as plt
import pickle
import time
import numpy as np

import ast
from matplotlib.colors import LinearSegmentedColormap
import matplotlib

from scipy.ndimage import gaussian_filter
from scipy.interpolate import interp1d
#to get a list of measurement filenames given:
# path, staticname eg.: 'S211026_', indexes: (1,200) and estension (.dat default) 
def getfnames(path,staticname, idx,extension='.dat'):
    fnames = []
    for i in idx:
        fnames.append(path + staticname + '{:03}'.format(i) + extension)
    return fnames

def getfnames_5(path,staticname, idx,extension='.dat'):
    fnames = []
    for i in idx:
        fnames.append(path + staticname + '{:05}'.format(i) + extension)
    return fnames

from scipy.signal import spectral
import deconvolution as deconv
import nanonis
spectra = nanonis.biasSpectroscopy()
#output the average of the conductance of list of filenames (spectra)
def avg_cond(filenames):
    conductance_avg = 0
    for f in filenames:
        spectra.load(f)
        spectra.normalizeRange([3e-3,4e-3])
        conductance_avg = conductance_avg + spectra.conductance
    conductance_avg = conductance_avg/len(filenames)
    return spectra.bias,conductance_avg

def avg_cond_decon(filenames):
    conductance_avg = 0
    for f in filenames:
        spectra.load(f)
        spectra.normalizeRange([3e-3,4e-3])
        bias, spectra.conductance = deconv.dynesDeconvolute_nof(spectra.bias,spectra.conductance,gap=1.4e-3,temperature=1.7, dynesParameter=40E-6, energyR=6E-3, spacing = 53.5e-6,x_min=-2.0E-3,x_max=2.0E-3,N=100)
        conductance_avg = conductance_avg + spectra.conductance
    conductance_avg = conductance_avg/len(filenames)
    return bias,conductance_avg/conductance_avg[0]


def set_size_cm(w,h, ax=None):
    """ w, h: width, height in cm """
    cm = 1/2.54
    if not ax: ax=plt.gca()
    l = ax.figure.subplotpars.left
    r = ax.figure.subplotpars.right
    t = ax.figure.subplotpars.top
    b = ax.figure.subplotpars.bottom
    figw = float(w*cm)/(r-l)
    figh = float(h*cm)/(t-b)
    ax.figure.set_size_inches(figw, figh)

def set_size(w,h, ax=None):
    """ w, h: width, height in inches """
    if not ax: ax=plt.gca()
    l = ax.figure.subplotpars.left
    r = ax.figure.subplotpars.right
    t = ax.figure.subplotpars.top
    b = ax.figure.subplotpars.bottom
    figw = float(w)/(r-l)
    figh = float(h)/(t-b)
    ax.figure.set_size_inches(figw, figh)

def save_obj(obj, name ):
    with open(name + '.pkl', 'wb') as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL) 
def load_obj(name ):
    with open( name + '.pkl', 'rb') as f:
        return pickle.load(f)


def didv(axs):
    if type(axs) == type(np.zeros(2)):
        for ax in axs:
            ax.set_xlabel('Bias (mV)',fontname="Arial")
            ax.set_ylabel('dI/dV '+r'(G$_N$)',fontname="Arial")
            ax.tick_params(axis='both',direction='in')
    else:
        axs.set_xlabel('Bias (mV)',fontname="Arial")
        axs.set_ylabel('dI/dV '+r'(G$_N$)',fontname="Arial")
        axs.tick_params(axis='both',direction='in')

def didv_p():
        plt.xlabel('Bias (mV)',fontname="Arial")
        plt.ylabel('dI/dV '+r'(G$_N$)',fontname="Arial")
        plt.tick_params(axis='both',direction='in')

def didv_dec(axs):
    if type(axs) == type(np.zeros(2)):
        for ax in axs:
            ax.set_xlabel('Bias (mV)')
            ax.set_ylabel('dI/dV dec. '+r'(G$_N$)')
            ax.tick_params(axis='both',direction='in')
    else:
        axs.set_xlabel('E-E'+r'$_F$ (meV)')
        axs.set_ylabel('dI/dV dec. '+r'(G$_N$)')
        axs.tick_params(axis='both',direction='in')

def inner(axs):
    if type(axs) == type(np.zeros(2)):
        for axn in axs:
            axn.tick_params(axis='x', direction='in')
            axn.tick_params(axis='y', direction='in')
    else:
            axs.tick_params(axis='x', direction='in')
            axs.tick_params(axis='y', direction='in')

def timing(i,t0,cycles):
    if i == 1:
        #to predict job time
        t1 = time.time()
        cycle = (t1-t0)/60
        total = np.round(cycle*cycles,3)
        if total <1:
            print('1 cycle:',np.round(cycle,3),'minutes --- Total time:',np.round(total*60,2),'seconds')
        else:
            print('1 cycle:',np.round(cycle,3),'minutes --- Total time:',total,'minutes')

def energyFind(bias, energy):
    index = (abs(bias - energy)).argmin()
    return index



def data_smooth(x,y,order=1):
    interp_func = interp1d(x, y, kind='cubic')
    new_x = np.linspace(x.min(),x.max(),2000)
    int_y = interp_func(new_x)
    yy = gaussian_filter(int_y,order)
    return new_x,yy



def cmap_fromLut(fname): # converts a .lut file given its path to a LinearSegmentedColormap
    # locate control points number for each color [blue,green,red] and store data
    control_points = []
    data = []
    with open(fname, "r") as f:
        for line in f:
            data.append(line.strip())
            if 'Number of Control Points' in line:
                split = line.split(':')
                control_points.append(int(split[1].strip()))
    # locate the
    location = []
    n=0
    for i in data:
        if 'Info' in i:
            location.append(n)
        n+=1
    # separate and store the control points
    points_B = data[location[0]+2:location[0]+2+control_points[0]]
    points_G = data[location[1]+2:location[1]+2+control_points[1]]
    points_R = data[location[2]+2:location[2]+2+control_points[2]]

    def convert(list): # to convert from string to array format
        A = []
        for i in list:
            A.append(ast.literal_eval(i.split(':')[1].strip()))
        A = np.array(A)
        sorted_indices = np.argsort(A[:,0])
        arr = A[sorted_indices]
        return arr

    points_B = convert(points_B)
    points_G = convert(points_G)
    points_R = convert(points_R)
    #invert the color map to match the original
    points_B[:,1] = 255-points_B[:,1]
    points_G[:,1] = 255-points_G[:,1]
    points_R[:,1] = 255-points_R[:,1]

    # Create the colormap
    cmap_lut = LinearSegmentedColormap('lut_colormap', {
        'red':   [(x/255.0, y/255.0, y/255.0) for x, y in points_R],
        'green': [(x/255.0, y/255.0, y/255.0) for x, y in points_G],
        'blue':  [(x/255.0, y/255.0, y/255.0) for x, y in points_B]
    })
    return cmap_lut


def export_colormap(colormap, filename):
    cmaplist = [colormap(i) for i in range(colormap.N)]
    cmaplist[0] = (1.0,1.0,1.0,1.0)
    colormap = matplotlib.colors.LinearSegmentedColormap.from_list('mcm',cmaplist, colormap.N)

    with open(filename, 'w') as file:
        file.write("WSxM file copyright UAM\n")
        file.write("New Format Palette. 2001\n")
        file.write("Image header size: 1143\n\n")
        
        def write_color_points(color_name, color_points):
            file.write(f"[{color_name} Info]\n")
            for i, point in enumerate(color_points):
                file.write(f"    Control Point {i}: ({int(point[0]*255)} , {255-int(point[1]*255)})\n")
            file.write(f"    Number of Control Points: {len(color_points)}\n\n")
        
        write_color_points("Blue", colormap._segmentdata['blue'])
        write_color_points("Green", colormap._segmentdata['green'])
        write_color_points("Red", colormap._segmentdata['red'])
        
        file.write("[Palette Generation Settings]\n")
        file.write("    Derivate Mode for the last blue Point: Automatic\n")
        file.write("    Derivate Mode for the last green Point: Automatic\n")
        file.write("    Derivate Mode for the last red Point: Automatic\n")
        file.write("    Is there a particular palette index colored?: No\n")
        file.write("    Smooth Blue: No\n")
        file.write("    Smooth Green: No\n")
        file.write("    Smooth Red: No\n\n")
        
        file.write("[Header end]")

import matplotlib
def spines(ax):
    plt.setp(ax.spines.values(), lw=0.3)
    ax.tick_params(width=0.3)
    new_rc_params = {'text.usetex': False,
    "svg.fonttype": 'none'
    }
    matplotlib.rcParams.update(new_rc_params)
    matplotlib.rcParams['axes.unicode_minus']=False


import shutil
import os

def copy_files(file_paths, destination_folder):
    """
    Copies files from a list of paths to a defined destination folder.

    :param file_paths: List of file paths to copy
    :param destination_folder: Destination folder where files will be copied
    """
    if not os.path.exists(destination_folder):
        os.makedirs(destination_folder)

    for file_path in file_paths:
        if os.path.isfile(file_path):
            try:
                shutil.copy(file_path, destination_folder)
                print(f"Copied {file_path} to {destination_folder}")
            except Exception as e:
                print(f"Error copying {file_path} to {destination_folder}: {e}")
        else:
            print(f"File not found: {file_path}")

from matplotlib.widgets import Slider
def explore(map):
    # Initial plot setup
    fig, ax = plt.subplots()
    plt.subplots_adjust(left=0.1, bottom=0.3)  # Adjust plot to make room for sliders

    # Initial vmax and vmin
    vmin = np.min(map)
    vmax = np.max(map)

    # Initial plot
    im = ax.imshow(map, aspect='auto', vmin=vmin, vmax=vmax, cmap='viridis')
    plt.colorbar(im, ax=ax)

    # Sliders for vmax and vmin
    axcolor = 'lightgoldenrodyellow'
    ax_vmin = plt.axes([0.1, 0.15, 0.65, 0.03], facecolor=axcolor)
    ax_vmax = plt.axes([0.1, 0.1, 0.65, 0.03], facecolor=axcolor)

    slider_vmin = Slider(ax_vmin, 'vmin', np.min(map), np.max(map), valinit=vmin)
    slider_vmax = Slider(ax_vmax, 'vmax', np.min(map), np.max(map), valinit=vmax)

    # Update function to be called when sliders are changed
    def update(val):
        im.set_clim([slider_vmin.val, slider_vmax.val])
        fig.canvas.draw_idle()

    # Connect the sliders to the update function
    slider_vmin.on_changed(update)
    slider_vmax.on_changed(update)

    plt.show()




# Superconductor analysis module
# nanoImaging group @ nanoGune

import numpy as np
from scipy import signal
from lmfit import Model
import nanonis
import matplotlib.pyplot as plt
import scipy.constants as const
from .Green_functions import spin5_sq
import deconvolution as decon

# Evaluate the non-broadened DOS of a conventional s-wave superconductor
#       * energy is the energy array or value
#       * Delta is the superconducting gap at 0K


# def BCS_curve(energy, Delta):
#     Delta = np.complex(Delta)
#     density = np.sign(energy)*np.real(np.divide(energy, np.sqrt(np.power(energy, 2)-np.power(Delta, 2))))
#     return density

# Evaluate the Dynes-broadened DOS of a conventional s-wave superconductor
#       * energy is the energy array
#       * Delta is the superconducting gap at 0K
#       * dynesParameter is the Dynes factor


def dynes_curve(energy,Delta,dynesParameter):
    dynesParameter = complex(0, dynesParameter)
    density = np.sign(energy)*np.real(np.divide(energy-dynesParameter,np.sqrt(np.power(energy-dynesParameter, 2)-np.power(Delta, 2))))
    return density

def dynes_javi(energy,dynesParameter,Delta):
    dynesParameter = complex(0, dynesParameter)
    density = np.sign(energy)*np.real(np.divide(energy-dynesParameter,np.sqrt(np.power(energy-dynesParameter, 2)-np.power(Delta, 2))))
    return density

def dynesdos(E, Gamma, Delta): #dynes function
    dos = np.real((np.abs(E+1j*Gamma))/np.sqrt((E+1j*Gamma)**2-Delta**2))
    return np.abs(dos)

def dynes_wrong(E, Delta, Gamma): #dynes function
    dos = np.real((E+1j*Gamma)/np.sqrt((E+1j*Gamma)**2-Delta**2))
    return np.abs(dos)/np.abs(dos)[0]


def fdd( E, mu, T): #fermi Dirac function
    if T == 0:
        f = np.heaviside(-(E-mu), 1)
    else:
        f = 1/(1+np.exp((E-mu)/(const.k*T/const.e)))
    return f


# Evaluate the derivative Dynes-broadened DOS of a conventional s-wave superconductor
#       * energy is the energy array
#       * Delta is the superconducting gap at 0K
#       * dynesParameter is the Dynes factor


def dynes_curve_diff(energy, Delta, dynesParameter):
    dynesParameter = np.complex(0, dynesParameter)
    density_de = -np.sign(energy)*np.real(np.divide(np.square(Delta),np.power(np.square(energy-dynesParameter)-np.square(Delta), 1.5)))
    density_de = np.nan_to_num(density_de)
    return density_de



# Evaluate Tinkham model for Coulomb blockade

def Ip(Vp,R2,C1,C2,Q0,delta=0):
    n0 = np.round(((-C2*Vp+Q0-const.e/2)/const.e + (-C2*Vp+Q0+const.e/2)/const.e)/2)
    if (-const.e/2 + n0*const.e -Q0 )/C1-delta < Vp <  (const.e/2 + n0*const.e -Q0)/C1+delta:
        Ip = 0
    else:
        Ip = (1/(R2*(C1+C2)))*(-(n0*const.e-Q0)+C1*Vp-const.e*np.sign(Vp)/2)
    return Ip

def fdd( E, mu, T): #fermi Dirac function
    if T == 0:
        f = np.heaviside(-(E-mu), 1)
    else:
        f = 1/(1+np.exp((E-mu)/(const.k*T/const.e)))
    return f



def coulomb(R2,C1,C2,Q0,Vpx=300,T=1.3,Erange=4e-3):
    V = np.linspace(-Erange,Erange,Vpx)
    Vint = np.linspace(-Erange/2,Erange/2,Vpx)
    I = []
    for Vp in V:
        I.append(Ip(Vp,R2,C1,C2,Q0))
    It = []
    for i in Vint:
        It.append(np.trapz(np.gradient(I)*( fdd(V,i,T)-fdd(V,0,T) ),x=V))
    return np.gradient(It)




# Fit a spectra with a dynes

def dynes_fit(fname, Delta=1e-3, dynes=1e-5):
    # initialize model
    dynes_Model = Model(dynes_curve)
    params = dynes_Model.make_params()
    params['Delta'].set(Delta, vary=True)
    params['dynesParameter'].set(dynes, vary=True)

    # load data
    spectra = nanonis.biasSpectroscopy()
    spectra.load(fname)
    x = np.array(spectra.bias)
    y = np.array(spectra.conductance)

    # perform fit
    result = dynes_Model.fit(y, energy=x, params=params)
    comps = result.eval(x=x)

    # visualization
    fig, ax = plt.subplots(1)
    ax.plot(x, y)
    ax.plot(x, comps)
    return result

# Fermi broadened dynes fit


class Superconductor():

    def __init__(self) -> None:
        pass

    def load(self,bias,conductance):
        self.conductance = conductance
        self.bias = bias

    def fdd(self, E, mu, T): #fermi Dirac function
        if T == 0:
            f = np.heaviside(-(E-mu), 1)
        else:
            f = 1/(1+np.exp((E-mu)/(const.k*T/const.e)))
        return f

    def dynesdos(self, E, Gamma, Delta): #dynes function
        dos = np.real((np.abs(E+1j*Gamma))/np.sqrt((E+1j*Gamma)**2-Delta**2))
        return np.abs(dos)


    def dynes_curve(self,energy,dynesParameter,Delta):
        dynesParameter = complex(0, dynesParameter)
        density = np.sign(energy)*np.real(np.divide(energy-dynesParameter,np.sqrt(np.power(energy-dynesParameter, 2)-np.power(Delta, 2))))
        return density

    def dynes_javi(self,energy,dynesParameter,Delta):
        dynesParameter = complex(0, dynesParameter)
        density = np.sign(energy)*np.real(np.divide(energy-dynesParameter,np.sqrt(np.power(energy-dynesParameter, 2)-np.power(Delta, 2))))
        return density

    def arnolddos(self,E,x=1,k=1.2e10,d=3.8e-7,Delta=0.75e-3,gamma=0.0000001):
        #atomic units constants
        a0 = 5.2917721e-11
        m_e = 0.510e6/const.c**2
        Ry_eV = const.hbar**2/(const.e**2*m_e*a0**2)
        #convert into atomic units
        Delta = Delta/Ry_eV
        d = d/a0
        k = k*a0
        x = -d
        #generate energy array
        E = E/Ry_eV +gamma*1j
        #green function compute
        m_e = 1
        kp = np.sqrt(k**2 + 2*m_e*E)
        km = np.sqrt(k**2  -2*m_e*E)
        FE = np.abs(E)/(np.sqrt(E**2-Delta**2))
        G =(m_e/(k))*((1j* FE* np.cos(kp*x+km*d) - np.sin(kp*x+km*d) )*( np.cos(kp*(x+d) ))/((1j*FE* np.sin((kp-km)*d)-np.cos((kp-km)*d)) ))
        return np.abs(np.imag(G))



    def SIS(self,bias, T, Delta1, Gamma1, Delta2, Gamma2,A): #General calculation of the dos with dynes in the tip and sample, broadened by Fermi Dirac
        curr = []
        self.E = np.linspace(-Delta1*10,Delta1*10,5000)
        
        for Vp in self.bias:
            currp = np.trapz(self.dynesdos(self.E, Gamma1, Delta1)*self.dynesdos(self.E-Vp,
                             Gamma2, Delta2)*(self.fdd(self.E, Vp, T)-self.fdd(self.E, 0, T)), x=self.E)
            curr.append(currp)
        out = A*np.gradient(np.array(curr))
        return out/out[0]
    
    def SISjavi(self,bias, T, Delta1, Gamma1, Delta2, Gamma2,A): #General calculation of the dos with dynes in the tip and sample, broadened by Fermi Dirac
        curr = []
        self.E = np.linspace(-Delta1*10,Delta1*10,5000)
        
        for Vp in self.bias:
            currp = np.trapz(self.dynes_javi(self.E, Gamma1, Delta1)*self.dynes_javi(self.E-Vp,
                             Gamma2, Delta2)*(self.fdd(self.E, Vp, T)-self.fdd(self.E, 0, T)), x=self.E)
            curr.append(currp)
        out = A*np.gradient(np.array(curr))
        return out/out[0]

    def dynesconv(self,bias,T,Delta,Gamma):
        curr = []
        for Vp in self.biasHR:
            currp = np.trapz(self.dynesdos(self.biasHR, Gamma, Delta)*(self.conductance/self.conductance[0])*(self.fdd(self.biasHR, Vp, T)-self.fdd(self.biasHR, 0, T)),x=self.biasHR)
            curr.append(currp)
        return np.gradient(np.array(curr))

    def dynes_arnold_conv(self,x,T,Delta_s,Delta_t,gamma_t,gamma_s=1,d=1,k=1.2e10):
        curr = []
        bias = np.linspace(-4*(Delta_t), 4*(Delta_t), 1000)
        E = np.linspace(bias[0],bias[-1],1000)
        for Vp in bias:
            currp = np.trapz(self.dynesdos(E, gamma_t, Delta_t)*(self.arnolddos(E-Vp,gamma=gamma_s,d=d,k=k,Delta=Delta_s))*(self.fdd(E, Vp, T)-self.fdd(E, 0, T)),x=E)
            curr.append(currp)
        out = -np.gradient(np.array(curr))
        return bias, out/out[0]


    def YSRconv(self,bias,J,delta_s,delta_t,gamma_s,gamma_t): #spectra of a single ysr convoluted with dynes
        m=20.956
        pf=0.274
        c = const.physical_constants['Hartree energy'][0]/const.e
        
        #simulate YSR
        sim = spin5_sq(2 , [J,J] , [0,0] , ((0,0),(12,6,6.3)) ,m=m,pf=pf,delta=delta_s/c)
        conductance = []
        E = np.linspace(-8*delta_s ,8*delta_s,900)

        for i in E/c:
            conductance.append(np.sign(i)*sim.DOS((0,0),i+gamma_s*1j*np.sign(i)/c))
        conductance = np.array(conductance)
        #convolution with tip
        bias = np.linspace(-4*delta_s,4*delta_s,300)
        conv_cond = decon.dynesConvolute(bias,E,conductance,delta_t,1.1,gamma_t)
        return conv_cond/conv_cond[0],bias

    def fitYSR(self):
        model = Model(self.YSRconv)
        params = model.make_params()
        params['J'].set(-0.0296,vary=False)
        params['delta_s'].set(0.71e-3,vary=False)
        params['delta_t'].set(1,vary=False,expr='delta_s')
        params['gamma_s'].set(60e-6,vary=False)
        params['gamma_t'].set(60e-6,vary=False)
        self.fit_res = model.fit(self.conductance,params,bias=self.bias)
        self.fit_res_eval = self.fit_res.eval(x=self.bias)


    def cond_spline():
        pass

    def fitSIS(self, T, Delta1, Gamma1, Delta2, Gamma2,A):
        model = Model(self.SIS)
        params = model.make_params()
        params['T'].set(T,vary=False)
        params['Delta1'].set(Delta1,vary=True)
        params['Gamma1'].set(Gamma1,vary=True)
        params['Delta2'].set(Delta2,vary=False)
        params['Gamma2'].set(Gamma2,vary=False)
        params['A'].set(A,vary=False)

        self.fit_res = model.fit(self.conductance,params,bias=self.bias)
        self.fit_res_eval = self.fit_res.eval(x=self.bias)
        return 
    
    def fit_arnold(self):
        model = Model(self.dynes_arnold_conv)
        self.params = model.make_params()
        self.params['T'].set(1,vary=False)
        self.params['Delta'].set(0.75e-3,vary=True)
        self.params['Gamma'].set(1e-6,min=0,vary=False)
        self.params['gamma'].set(4e-6,min=0,vary=True)
        self.params['d'].set(8e-5,vary=True)
        self.params['k'].set(3.2e12,vary=False)
        self.fit_res = model.fit(self.conductance,self.params,x=self.bias)
        self.fit_res_eval = self.fit_res.eval(x=self.bias)


    def showResults(self):
        fig,self.ax = plt.subplots(1)
        self.ax.plot(self.bias,self.conductance)
        self.ax.plot(self.bias,self.fit_res.eval(x=self.bias))

    def dec_Cut(self,Energy):
        idx_n = abs(self.bias+Energy).argmin()
        idx_p = abs(self.bias-Energy).argmin()
        self.bias = self.bias[idx_p:idx_n]
        self.conductance = self.conductance[idx_p:idx_n]



class Arnold(): #compute Arnold model DOS in atomic units

    def __init__(self):
        pass

    def G(self,x=1,Erange=1,k=1,d=1,Delta=1,gamma=1,Epx=10000):
        #atomic units constants
        self.a0 = 5.2917721e-11
        m_e = 0.510e6/const.c**2
        self.Ry_eV = const.hbar**2/(const.e**2*m_e*self.a0**2)
        #convert into atomic units
        Delta = Delta/self.Ry_eV
        d = d/self.a0
        k = k*self.a0
        x = x/self.a0
        #generate energy array
        self.E= np.linspace(-Erange/self.Ry_eV,Erange/self.Ry_eV,Epx) +gamma*1j
        self.Er = np.real(self.E)
        #green function compute
        E = self.E
        m_e = 1
        kp = np.sqrt(k**2 + 2*m_e*E)
        km = np.sqrt(k**2  -2*m_e*E)
        FE = np.abs(E)/(np.sqrt(E**2-Delta**2))
        G =(m_e/(k))*((1j* FE* np.cos(kp*x+km*d) - np.sin(kp*x+km*d) )*( np.cos(kp*(x+d) ))/((1j*FE* np.sin((kp-km)*d)-np.cos((kp-km)*d)) ))
        return np.abs(np.imag(G))




from pandas.core.reshape.tile import cut
from . import nanonis
import matplotlib.pyplot as plt
import matplotlib.gridspec as gs
from matplotlib import rc, ticker
from matplotlib.widgets import Button
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar
import numpy as np
from scipy import ndimage
import csv
import glob as glob
from matplotlib.widgets import Slider
import os
import pandas as pd
import colorcet as cc
from lmfit import Model
from scipy import stats
import useful as uf
#added: now the linescans cuts save properly in order
#       new class for the lineprofile cuts, given a folder with cuts it plot them
#       new class for plotting general image with sliders


#lineProfile() is a class dependent on nanonis.linescan(). Adds the graphing features:
# if the plugins are on: sliders for vmax vmin, vertical cut with or without the influence radius
# cutmode different than line allows to click and retrieve the point number saved in self.points
# flipud='on' flips the map up-down
# noline='on' is for loading datasets that are not linescans but we still want to plot them with imshow.
# LStype change allows to load different LS formats. Can be: 'normal', the conventional set of .dat files, 'wsxm' a txt file with a LS extracted from WSXM, .3ds a linescan saved in binary.
class lineProfile():

    def __init__(self,spac=0,categorical='python',plotmode='cmap',influence='off',range=0.1e-3,plugins='on',cutMode='line',flipud='off',LStype='normal'): # vmin/vmax colourscale, cut=True enables vertical cuts
        #import the settings for the cmap plotting
        self.cutMode = cutMode
        self.LStype = LStype
        self.range = range
        self.colormap = 'YlGnBu_r'
        self.influence = influence
        self.flipud= flipud
        #import settings for cascade plotting
        self.plotmode, self.spac, self.categorical = plotmode,spac, categorical


        self.figure = plt.figure(figsize = (5,3))
        if plugins=='on': # enables the cmap changes 
            self.figure.subplots_adjust(bottom=0.3)
            grid = gs.GridSpec(2, 1, height_ratios=[2, 1])
            self.axMap = self.figure.add_subplot(grid[0])
            self.axCut = self.figure.add_subplot(grid[1])
            self.axCut.set_xlabel('Distance (nm)')
            self.axCut.set_ylabel('dI/dV (arb. units)')
            self.figure.canvas.mpl_connect('button_press_event', self.mapClick)
            self.axmin = self.figure.add_axes([0.25, 0.1, 0.65, 0.03])
            self.axmax = self.figure.add_axes([0.25, 0.15, 0.65, 0.03])
            self.smin = Slider(self.axmin, 'Min', -4, 8, valinit =0)
            self.smax = Slider(self.axmax, 'Max', -4, 8, valinit =4)
            self.smin.on_changed(self.update)
            self.smax.on_changed(self.update)
        if plugins=='off':
            self.axMap = self.figure.add_subplot(111)
        self.figure.show()

    def draw(self):
        if self.plotmode == 'cmap':
            self.im1 = self.axMap.imshow(self.linescan.conductance, aspect='auto', extent=self.extent,cmap = self.colormap, interpolation='nearest', vmin=self.vmin, vmax=self.vmax)
        if self.plotmode == 'cascade':
            if self.categorical == 'viridis':
                N = self.linescan.conductance.shape[0]
                plt.rcParams["axes.prop_cycle"] = plt.cycler("color", plt.cm.viridis(np.linspace(0,1,N)))
            for i in range(0,N):
                self.axMap.plot(self.linescan.bias*1e3,self.linescan.conductance[i,:]+i*self.spac)

    def mapLoad(self,filenames,distance=None): # distance is the path of the distance file for wsxm LS type
        if self.LStype == 'normal': #for linescans recorded in series of .dat files
            self.linescan = nanonis.linescan()
            self.linescan.load(filenames)
            self.axMap.set_title(self.linescan.name[0]+' - '+self.linescan.name[-1], fontweight='bold')
            self.cmin = self.linescan.conductance.min()
            self.cmax = self.linescan.conductance.max()
            self.vmin = self.linescan.conductance.min()
            self.vmax = self.linescan.conductance.max()
            self.axMap.set_ylabel("Distance (nm)")
            self.axMap.set_xlabel("Bias (mV)")
            plt.subplots_adjust(hspace=0.35)
            self.extent = [min(self.linescan.bias*1e3), max(self.linescan.bias*1e3), min(self.linescan.distance), max(self.linescan.distance)]
            if self.flipud=='on':
                self.linescan.conductance = np.flipud(self.linescan.conductance)
            self.draw()
        if self.LStype == '3ds': #checks if we load a 3ds files instead of ascii files
            self.linescan = nanonis.linescan3ds()
            self.linescan.load(filenames)
            self.axMap.set_title(self.linescan.name)
            self.vmin = self.linescan.conductance.min()
            self.vmax = self.linescan.conductance.max()
            self.cmin = self.linescan.conductance.min()
            self.cmax = self.linescan.conductance.max()
            self.axMap.set_ylabel("Distance (nm)")
            self.axMap.set_xlabel("Bias (mV)")
            plt.subplots_adjust(hspace=0.35)
            self.extent = [min(self.linescan.bias*1e3), max(self.linescan.bias*1e3), min(self.linescan.distance), max(self.linescan.distance)]
            self.draw()
        if self.LStype == 'wsxm': #for linescans extracted from wsxm, need a distance file path
            self.linescan = nanonis.linescan()
            self.linescan.conductance,self.linescan.bias,self.linescan.distance = nanonis.readGcutLS(filenames,distance)
            self.linescan.name = filenames
            self.axMap.set_title(filenames)
            self.cmin = self.linescan.conductance.min()
            self.cmax = self.linescan.conductance.max()
            self.vmin = self.linescan.conductance.min()
            self.vmax = self.linescan.conductance.max()
            self.axMap.set_ylabel("Distance (nm)")
            self.axMap.set_xlabel("Bias (mV)")
            plt.subplots_adjust(hspace=0.35)
            self.extent = [min(self.linescan.bias), max(self.linescan.bias), min(self.linescan.distance), max(self.linescan.distance)]
            self.draw()

    def mapCut(self,range=[-2e-3,2e-3]):
        idxs = [abs(self.linescan.bias-range[0]).argmin(),abs(self.linescan.bias-range[1]).argmin()]
        print(idxs)
        self.linescan.conductance = self.linescan.conductance[:,idxs[-1]:idxs[0]]
        self.linescan.bias = self.linescan.bias[idxs[-1]:idxs[0]]
        self.extent = [min(self.linescan.bias*1e3), max(self.linescan.bias*1e3), min(self.linescan.distance), max(self.linescan.distance)]
  
    def mapScale(self, min, max):
        self.cmin = min
        self.cmax = max
        self.axMap.cla()
        self.draw()

    def mapColor(self, mapCol):
        self.colormap = mapCol
        self.axMap.cla()
        self.draw()
    
    def mapGaussian(self, radius):
        self.linescan.conductance = ndimage.gaussian_filter(self.linescan.conductance, radius)
        self.axMap.cla()
        self.draw()

    def distanceOffset(self, offset):
        self.linescan.distanceOffset(offset)
        self.axMap.cla()
        self.draw()

    def normalize_offset(self, energy, offset):
        self.linescan.normalizeTo(energy)
        self.linescan.biasOffset(offset)
        self.axMap.cla()
        self.axMap.set_title(self.linescan.name[0]+' - '+self.linescan.name[-1], fontweight='bold')
        self.cmin = self.linescan.conductance.min()
        self.cmax = self.linescan.conductance.max()
        self.axMap.set_ylabel("Distance (nm)")
        self.axMap.set_xlabel("Bias (mV)")
        plt.subplots_adjust(hspace=0.35)
        self.draw()
    
    def normalize_range(self,E_range,offset):
        self.linescan.normalizeRange(E_range)
        self.linescan.biasOffset(offset)
        self.axMap.cla()
        self.axMap.set_title(self.linescan.name[0]+' - '+self.linescan.name[-1], fontweight='bold')
        self.cmin = self.linescan.conductance.min()
        self.cmax = self.linescan.conductance.max()
        self.axMap.set_ylabel("Distance (nm)")
        self.axMap.set_xlabel("Bias (mV)")
        plt.subplots_adjust(hspace=0.35)
        self.draw()

    def mapClick(self,event):
        if self.cutMode == 'point':
            if event.inaxes == self.axMap:
                if event.dblclick: #cancel the last tracked point
                    self.YSRtrace[0] = self.YSRtrace[0][:-2]
                    self.YSRtrace[1] = self.YSRtrace[1][:-2] 
                elif self.influence == 'off':
                    self.energyCut = event.xdata
                    self.distCut = event.ydata
                    self.points = [[],[]]
                    self.cutSave(self.energyCut,self.distCut)
                else:
                    self.energyCut = event.xdata
                    self.distCut = event.ydata
                    self.cutPlotRange(self.energyCut,self.distCut)
        elif self.cutMode == 'line':
            if event.inaxes == self.axMap:
                if event.dblclick:
                    self.spectraIndex = event.y 
                elif self.influence == 'off':
                    self.energyCut = event.xdata
                    self.cutPlot(self.energyCut*1e-3)
                else:
                    self.energyCut = event.xdata
                    self.cutPlotRange(self.energyCut*1e-3)
                    self.distCut = event.ydata
                    self.points = [[],[]]
                    self.cutSave(self.energyCut,self.distCut)         
    
    def cutSave(self,energy,distCut): #save the point clicked on the plot and put a dot in the graph
        id = (abs(self.linescan.bias-energy)).argmin()
        id_d = (abs(self.linescan.distance-distCut)).argmin()
        self.axMap.scatter(self.linescan.bias[id],self.linescan.distance[id_d])
        self.points[0].append(self.linescan.distance[id_d])
        self.points[1].append(self.linescan.bias[id])
        self.saveCSV(self.linescan.distance, self.linescan.conductance[:,id])
        self.figure.canvas.draw_idle()

    def cutPlot(self, energy):
        id = (abs(self.linescan.bias-energy)).argmin()
        self.axMap.plot([self.linescan.bias[id]*1e3,self.linescan.bias[id]*1e3],[self.linescan.distance[0],self.linescan.distance[-1]])
        self.axCut.plot(self.linescan.distance, self.linescan.conductance[:,id], label=str(round(self.linescan.bias[id]*1e3,2)))
        self.axCut.legend()
        self.saveCSV(self.linescan.distance, self.linescan.conductance[:,id])
        self.figure.canvas.draw_idle()
    
    def cutPlotRange(self, energy):
        def LSinfluence_avg(id_c,id_n,id_p): #return contuctance averaged between different cuts
            idxs = np.arange(id_p,id_n)
            LSconductance_avg = np.zeros(len(self.linescan.conductance))
            if id_n == id_p: #checks that the range is >0
                LSconductance_avg = self.linescan.conductance[:,id_c]
            else:
                for idx in idxs:
                    LSconductance_avg = LSconductance_avg + self.linescan.conductance[:,idx]
                LSconductance_avg = LSconductance_avg/len(idxs)
            return LSconductance_avg
        
        #calculate the index based on the range given and the energy
        id_c = (abs(self.linescan.bias-energy)).argmin()
        id_n = (abs(self.linescan.bias-energy-self.range)).argmin()
        id_p = (abs(self.linescan.bias-energy+self.range)).argmin()
        self.conductance_avg = LSinfluence_avg(id_c,id_n,id_p)
        self.axMap.fill_between([self.linescan.bias[id_n]*1e3,self.linescan.bias[id_p]*1e3],self.linescan.distance[0],self.linescan.distance[-1],alpha=0.6)
        self.axCut.plot(self.linescan.distance,self.conductance_avg,label=str(round(self.linescan.bias[id_c]*1e3,4)))
        self.axCut.legend()
        # self.saveCSV(self.linesan.distance, self.conductance_avg)
        self.figure.canvas.draw_idle()
         
    def saveCSV(self, array1,array2):
        #for incremental save
        stridx = self.linescan.name[0].find('LS')
        LSidx = self.linescan.name[0][stridx:stridx+4]
        Eidx = np.round(self.energyCut,2)
        count = 0
        for i in os.listdir():
            if 'Cut{}'.format(count) in i:
                count += 1
        matrix = np.vstack((array1,array2))
        matrix = np.transpose(matrix)
        filename_mod = "Cut{}_{}_{}.txt".format(count,LSidx,Eidx)
        with open(filename_mod, 'w') as csvfile:
            writer = csv.writer(csvfile)
            [writer.writerow(r) for r in matrix]


    def update(self, val): #for the color scale sliders
        self.im1.set_clim([self.smin.val,self.smax.val])
        self.figure.canvas.draw()

    
    def stepfinder(self,delta,cut,change): #to correct the step caused by the rolf divider
        self.offset_steps_a = np.zeros(5)
        self.offset_steps_b = np.zeros(5)
        
        deltaIdx = self.linescan.energyFind(delta)
        cutIdx = self.linescan.energyFind(cut)
        count_a = 0
        count_b = 0
        diff = abs(self.linescan.conductance[0,deltaIdx]-self.linescan.conductance[0,0])
        print(diff)
        for i in range(1, self.linescan.conductance.shape[0]):
            if self.linescan.conductance[i,cutIdx] - self.linescan.conductance[i-1,cutIdx] > diff*change:
                self.offset_steps_a[count_a] = int(i)
                count_a +=1
        for i in range(1, self.linescan.conductance.shape[0]):
            if self.linescan.conductance[i-1,cutIdx] - self.linescan.conductance[i,cutIdx] > diff*change:
                self.offset_steps_b[count_b] = int(i)
                count_b +=1
            self.offset_steps_a = self.offset_steps_a.astype(int)
            self.offset_steps_b = self.offset_steps_b.astype(int)
        return self.offset_steps_a, self.offset_steps_b 

    def random_offset(self, offset=0, delta=0,cut=0,change=0):
        self.stepfinder(delta,cut,change)
        print(self.stepfinder(delta,cut,change))
        energy_px = self.linescan.bias[0] - self.linescan.bias[1]
        offset_px = int(offset/energy_px)
        for i in range (0,len(self.offset_steps_a)):
            if self.offset_steps_b[i] == 0:
                self.linescan.conductance[self.offset_steps_a[i]:-1,:] = np.roll(self.linescan.conductance[self.offset_steps_a[i]:-1,:], -offset_px)
                break
            self.linescan.conductance[self.offset_steps_a[i]:self.offset_steps_b[i],:] = np.roll(self.linescan.conductance[self.offset_steps_a[i]:self.offset_steps_b[i],:], -offset_px)
            print('rolled')

    def hand_normalization(self,path_values,offset):
        values = pd.read_csv(path_values,header=None)
        values = values.to_numpy()
        for i in range(len(self.linescan.name)):
            self.linescan.conductance[i][:] = self.linescan.conductance[i][:]/values[i]
        self.linescan.biasOffset(offset)
        self.axMap.cla()
        self.axMap.set_title(self.linescan.name[0]+' - '+self.linescan.name[-1], fontweight='bold')
        self.cmin = self.linescan.conductance.min()
        self.cmax = self.linescan.conductance.max()
        self.axMap.set_ylabel("Distance (nm)")
        self.axMap.set_xlabel("Bias (mV)")
        plt.subplots_adjust(hspace=0.35)
        self.draw()

    def deconvolution(self,gap=1.37e-3, temperature=1.3, dynesParameter=40e-6, energyR=8e-3, spacing=35e-6,x_min=-4E-3,x_max=4E-3,N=300, window=15,order=2,n=2000):
        self.linescan.deconvolution(gap, temperature, dynesParameter, energyR, spacing,x_min,x_max,N, window,order,n)
        self.linescan.bias = np.flip(self.linescan.bias_dec) # bias flip cause in graphix the standard is from positive to negative
        self.linescan.conductance = self.linescan.conductance_dec

class multilineprofile():
    def LSload(self, start, end ,datestamp, path):
        filenames = []
        self.nLS = end - start + 1
        self.start = start
        self.end = end
        for i in range(start, end + 1):
            if i <10:
                textchunk = 'LS0' + str(i)
            else: 
                textchunk = 'LS' + str(i)
            filenames.append(glob.glob(path + textchunk + datestamp))
        return filenames

    def axis_setup(self): # removes the axis exept the bottom ones and apply add labels
        all_axes = self.figure.get_axes()
        i = self.start
        for ax in all_axes:
            ax.annotate(str(i) + ' At', (1,1))
            ax.set_yticks([])
            i += 1
            if ax.is_last_row():
                pass
            else: 
                ax.set_xticks([])

    def plotSetup(self, ): #plot settings
        fullwith = False
        if fullwith == True:
            self.extent=[min(self.linescan.bias), max(self.linescan.bias), min(self.linescan.distance), max(self.linescan.distance)]
        else: 
            self.extent = [-self.Ecut ,self.Ecut ,min(self.linescan.distance), max(self.linescan.distance)]
        self.cmin = 0
        self.cmax = 4

    def imgslice(self, Ecut): #cut the conductance at the given energy
        self.Ecut = Ecut
        idx = (np.abs(self.linescan.bias - Ecut)).argmin()
        self.linescan.conductance = self.linescan.conductance[:,idx:300-idx]
        return self.linescan.conductance

    def multimapdraw(self, filenames,Ecut):
        self.figure = plt.figure(figsize = (10,5))
        self.grid = gs.GridSpec(4, self.nLS // 5 + 1, wspace=0.0, hspace=0.0)
        self.figure.show()
        for i in range(0,len(filenames)):
            self.linescan = nanonis.linescan()
            self.linescan.load(filenames[i])
            self.axMap = self.figure.add_subplot(self.grid[i])
            self.linescan.conductance = np.fliplr(self.linescan.conductance)
            self.linescan.conductance = self.imgslice(Ecut) #slice the array
            self.plotSetup()
            self.axMap.imshow(self.linescan.conductance, aspect='auto', extent=self.extent, vmin=self.cmin,vmax=self.cmax)
            self.axMap.set_xlabel("Bias (mV)")
        self.axis_setup()

class map():

    def __init__(self):
        self.figure = plt.figure(figsize = (7,15))
        grid = gs.GridSpec(2, 1, height_ratios=[2, 2])
        self.axMap = self.figure.add_subplot(grid[0])
        self.axCut = self.figure.add_subplot(grid[1])
        #self.figure.canvas.mpl_connect('button_press_event', self.mapClick)
        fft_button_ax = self.figure.add_axes([0.65, 0.025, 0.1, 0.04])
        self.fft_button = Button(fft_button_ax, 'FFT', hovercolor='g')
        self.fft_button.on_clicked(self.fftClick)
        reset_button_ax = self.figure.add_axes([0.8, 0.025, 0.1, 0.04])
        self.reset_button = Button(reset_button_ax, 'Reset', hovercolor='g')
        self.reset_button.on_clicked(self.resetClick)
        self.figure.show()

        
    def channPlot(self, file, channel='Z', processing='Raw', title=[], gaussian=0, colormap='pink'):
        for j in range(len(file.data.keys())):
            channs = list(file.data.keys())
            if channel in channs[j]:
                self.map = file.data[channs[j]]
                if processing == 'firstPoint':
                    for i in range(self.map.shape[1]):
                        self.map[i][:] = self.map[i][:]-self.map[i][0]
                if processing == 'points':
                    for i in range(self.map.shape[1]):
                        self.map[i][:] = self.map[i][:]-np.average(self.map[i][:20])
                if processing == 'average':
                    for i in range(self.map.shape[1]):
                        self.map[i][:] = self.map[i][:]-np.average(self.map[i][:])
                self.map = ndimage.gaussian_filter(self.map, gaussian)
                if title:
                    self.axMap.set_title('{a} {b}\n{c} meV'.format(a=file.name, b=channs[j], c=title), fontweight='bold')
                else:
                    self.axMap.set_title('{a} {b}'.format(a=file.name, b=channs[j]), fontweight='bold')
                self.map = self.map-np.min(self.map)
                self.axMap.imshow(self.map*1e9, extent=[0, file.xrange, 0, file.yrange], interpolation='mitchell', cmap=colormap)
                break

    def resetClick(self, event):
        self.axCut.cla()

    def fftClick(self, event):
        fft = np.fft.fft2(self.map)
        fft = fft+np.rot90(np.rot90(fft))
        fft = abs(np.fft.fftshift(fft))
        self.axCut.imshow(fft, cmap='magma_r',vmax=10e-8)

class lineprofileCuts():
    def __init__(self,directory,delimiter=None):
        #read a path and return the filenames of the cuts
        self.cuts_path = []
        for i in os.listdir(directory):
            if '.txt' in i:
                self.cuts_path.append(i)
        #pack the cuts in a dataframe
        dfs = []
        for i in self.cuts_path:
            df = 0
            df = pd.read_csv(directory + '/' + i,delimiter=delimiter,header=None,names=[i +'_d',i +'_I'])
            dfs.append(df)
        self.df = pd.concat(dfs,axis=1)


    def plot_cuts(self,vertOff=0,xOffset=0): #plot the cuts with a vertical offset
        self.fig, ax = plt.subplots(1)
        count = 0
        for i in range(len(self.df.columns)):
            if count==len(self.df.columns):
                break
            ax.plot(self.df[self.df.columns[count]]+xOffset, self.df[self.df.columns[count+1]]+ count*vertOff, label=self.df.columns[count])
            count += 2
            xOffset=0 #only the first spectra is going to be offsetted
        ax.set_xlabel('Distance(nm)')
        ax.set_ylabel('dI/dV')
        plt.legend()
        plt.show()
        return
    def savefig(self,name):
        self.fig.savefig('{}.pdf'.format(name))

class LScut():
    def __init__(self,side='positive',center=15,ylim=None):
        self.side = side
        self.center = center
        self.ylim = ylim

    def load(self,filename,signal):#load the data
        self.rawdata = pd.read_csv(filename,delimiter=',',header=0)
        self.data = pd.DataFrame()
        self.data[0] = self.rawdata['d']
        self.data[1] = self.rawdata[signal]
        self.shiftToZero()
        self.centerToZero()
        #define new r for plotting the fit
        self.x = np.linspace(self.data[0].iloc[0],self.data[0].iloc[-1],1000)
        if self.side == 'positive': #to avoid root of negative number
            self.x = self.x[500:]
        elif self.side == 'negative':
            self.x = self.x[:500]

    def plot1side(self):
        self.fig,self.ax = plt.subplots(1)
        self.ax.clear()
        self.ax.plot(self.data[0],self.data[1])
        self.ax.set_xlabel('Distance (nm)')
        self.ax.set_ylabel('dI/dV (mV)')
        self.ax.set_ylim(-0.2,max(self.data[1]))
        if self.side == 'positive':
            self.ax.set_xlim(0,15)
        elif self.side == 'negative':
            self.ax.set_xlim(-15,0)
        
        #sliders 1st oscillation
        self.fig.subplots_adjust(bottom=0.5)
        self.i = 0
        self.t = 0.7
        self.c1 = 23
        self.p1 = 1.5
        self.A1 = 10
        self.k1 = 1.046
        self.axi = plt.axes([0.15, 0.07, 0.65, 0.03])
        self.axc1 = plt.axes([0.15, 0.1, 0.65, 0.03])
        self.axp1 = plt.axes([0.15, 0.13, 0.65, 0.03])
        self.axA1 = plt.axes([0.15, 0.16, 0.65, 0.03])
        self.axk1 = plt.axes([0.15, 0.19, 0.65, 0.03])
        self.axt = plt.axes([0.15, 0.22, 0.65, 0.03])
        self.is1 = Slider(self.axi, 'i',  0, 30, valinit = 0)
        self.cs1 = Slider(self.axc1, 'c1',  0, 30, valinit = 23)
        self.ps1 = Slider(self.axp1, 'p1', 0, 10, valinit = 1.5)
        self.As1 = Slider(self.axA1, 'A1', 0, 20, valinit = 10)
        self.ks1 = Slider(self.axk1, 'k1', 0.5, 10, valinit = 1.046)
        self.ts = Slider(self.axt,'t', 0, 3, valinit = 0.7)
        self.is1.on_changed(self.update)
        self.cs1.on_changed(self.update)
        self.ps1.on_changed(self.update)
        self.As1.on_changed(self.update)
        self.ks1.on_changed(self.update)
        self.ts.on_changed(self.update)
        #sliders 2nd oscillation
        self.c2 = 23
        self.p2 = 1.5
        self.A2 = 10
        self.k2 = 4.48
        self.axc2 = plt.axes([0.15, 0.25, 0.65, 0.03])
        self.axp2 = plt.axes([0.15, 0.28, 0.65, 0.03])
        self.axA2 = plt.axes([0.15, 0.31, 0.65, 0.03])
        self.axk2 = plt.axes([0.15, 0.34, 0.65, 0.03])
        self.cs2 = Slider(self.axc2, 'c2',  0, 30, valinit =23)
        self.ps2 = Slider(self.axp2, 'p2', 0, 10, valinit =1.5)
        self.As2 = Slider(self.axA2, 'A2', 0, 100, valinit =10)
        self.ks2 = Slider(self.axk2, 'k2', 1, 15, valinit =4.48)
        self.cs2.on_changed(self.update)
        self.ps2.on_changed(self.update)
        self.As2.on_changed(self.update)
        self.ks2.on_changed(self.update)
        #do the plotting of the fitting functions
        if self.side == 'positive':
            y = self.oscFuncpos1(self.x,self.A1,self.A2,self.k1,self.k2,self.p1,self.p2,self.c1,self.c2,self.t,self.i)
            self.ax.plot(self.x,y)
        elif self.side == 'negative':
            y = self.oscFuncneg1(self.x,self.A1,self.A2,self.k1,self.k2,self.p1,self.p2,self.c1,self.c2,self.t,self.i)
            self.ax.plot(self.x,y)
    
    def plotFull(self):
        self.fig,self.ax = plt.subplots(1)
        self.ax.plot(self.data[0],self.data[1])
        self.ax.set_xlabel('Distance (nm)')
        self.ax.set_ylabel('dI/dV')


    def draw1D(self): #
        self.ax.clear()
        self.ax.plot(self.data[0],self.data[1])
        self.ax.set_xlabel('Distance (nm)')
        self.ax.set_ylabel('dI/dV (mV)')
        self.ax.set_ylim(-0.2,self.ylim)
        if self.side == 'positive':
            self.ax.set_xlim(0,15)
        elif self.side == 'negative':
            self.ax.set_xlim(-15,0)

    def shiftToZero(self):
        min = np.min(self.data[1])
        self.data[1] = self.data[1]-min

    def centerToZero(self):
        self.data[0] = self.data[0]-self.center

    def maskcenter(self,cutOff):
        p_idx = abs(self.data[0]-cutOff).argmin()
        n_idx = abs(self.data[0]+cutOff).argmin()

        for i in self.data[0]:
            if np.abs(i) < cutOff:
                self.data.iloc[n_idx:p_idx,1] = np.mean(self.data[1])

    def oscFuncpos1(self,x,A1,A2,k1,k2,p1,p2,c1,c2,t,i):
        oscFunc1 = (A1*(np.sin(k1*x-p1))*np.exp(-x/c1))/(np.power(x,t)*k1)
        oscFunc2 = (A2*(np.sin(k2*x-p2))*np.exp(-x/c2))/(np.power(x,t)*k2)
        return oscFunc1**2 + oscFunc2**2 + i*oscFunc1*oscFunc2
    
    def oscFuncneg1(self,x,A1,A2,k1,k2,p1,p2,c1,c2,t,i):
        oscFunc1 = (A1*(np.sin(-k1*x-p1))*np.exp(x/c1))/(np.power(-x,t)*k1)
        oscFunc2 = (A2*(np.sin(-k2*x-p2))*np.exp(x/c2))/(np.power(-x,t)*k2)
        return oscFunc1**2 + oscFunc2**2 + i*oscFunc1*oscFunc2


    def update(self, val): #for the color scale sliders
        self.c1 = self.cs1.val
        self.p1 = self.ps1.val
        self.A1 = self.As1.val
        self.k1 = self.ks1.val
        self.c2 = self.cs2.val
        self.p2 = self.ps2.val
        self.A2 = self.As2.val
        self.k2 = self.ks2.val
        self.t = self.ts.val
        self.i = self.is1.val
        self.draw1D()
        if self.side == 'positive':
            y = self.oscFuncpos1(self.x,self.A1,self.A2,self.k1,self.k2,self.p1,self.p2,self.c1,self.c2,self.t,self.i)
            self.ax.plot(self.x,y)
        elif self.side == 'negative':
            y = self.oscFuncneg1(self.x,self.A1,self.A2,self.k1,self.k2,self.p1,self.p2,self.c1,self.c2,self.t,self.i)
            self.ax.plot(self.x,y)

    def cutData(self,distCut):
        if self.side == 'positive':
            idx = abs(self.data[0]-distCut).argmin()
            self.cuttedData = self.data.iloc[idx:,:]
        if self.side == 'negative':
            idx = abs(self.data[0]-distCut).argmin()
            self.cuttedData = self.data.iloc[:idx,:]
        return

    def autoFit(self):
        if self.side == 'positive':
            model = Model(self.oscFuncpos1)
            params = model.make_params()
            params['A1'].set(self.A1,vary=True)
            params['A2'].set(self.A2,vary=True)
            params['k1'].set(self.k1,vary=False)
            params['k2'].set(self.k2,vary=True)
            params['c1'].set(self.c1,vary=False)
            params['c2'].set(self.c2,vary=False)
            params['p1'].set(self.p1,vary=True)
            params['p2'].set(self.p2,vary=True)
            params['t'].set( self.t ,vary=False)
            params['i'].set( self.i ,vary=False)
        if self.side == 'negative':
            model = Model(self.oscFuncneg1)
            params = model.make_params()
            params['A1'].set(self.A1,vary=True)
            params['A2'].set(self.A2,vary=True)
            params['k1'].set(self.k1,vary=False)
            params['k2'].set(self.k2,vary=True)
            params['c1'].set(self.c1,vary=False)
            params['c2'].set(self.c2,vary=False)
            params['p1'].set(self.p1,vary=True)
            params['p2'].set(self.p2,vary=True)
            params['t'].set( self.t ,vary=False)
            params['i'].set( self.i ,vary=True)
        self.fitResult = model.fit(self.cuttedData[1],x=self.cuttedData[0],params=params)

        return 

    def plotfitResults(self):
        plt.figure()
        plt.plot(self.cuttedData[0],self.cuttedData[1])
        xp = np.linspace(self.cuttedData.iloc[0,0],self.cuttedData.iloc[-1,0],1000)
        comps = self.fitResult.eval(x=xp)
        plt.plot(xp,comps)

class Zapproach(nanonis.Zapproach):

    def __init__(self,vmin=0, vmax=4,influence='off',rangecut=0,plugins='on'):
        self.vmax = vmax
        self.vmin = vmin
        self.rangecut = rangecut
        self.colormap = 'YlGnBu_r'
        self.influence = influence
        self.plugins = plugins
        pass

    def draw(self):
        self.im1 = self.axMap.imshow(self.conductance, aspect='auto', extent=self.extent,cmap = self.colormap, interpolation='nearest', vmin=self.vmin, vmax=self.vmax)

    def mapload(self,fnames):
        self.load(fnames)

    
    def normalize_range(self,E_range):
        #initialize figure
        self.figure = plt.figure(figsize = (5,3))
        if self.plugins=='on':
            self.figure.subplots_adjust(bottom=0.3)
            grid = gs.GridSpec(2, 1, height_ratios=[2, 1])
            self.axMap = self.figure.add_subplot(grid[0])
            self.axCut = self.figure.add_subplot(grid[1])
            self.axCut.set_xlabel('Distance (nm)')
            self.axCut.set_ylabel('dI/dV (arb. units)')
            self.figure.canvas.mpl_connect('button_press_event', self.mapClick)
            self.axmin = self.figure.add_axes([0.25, 0.1, 0.65, 0.03])
            self.axmax = self.figure.add_axes([0.25, 0.15, 0.65, 0.03])
            self.smin = Slider(self.axmin, 'Min', -4, 8, valinit =0)
            self.smax = Slider(self.axmax, 'Max', -4, 8, valinit =4)
            self.smin.on_changed(self.update)
            self.smax.on_changed(self.update)
        if self.plugins=='off':
            self.axMap = self.figure.add_subplot(111)

        
        #load map
        self.normalizeRange(E_range)
        self.extent = [min(self.bias*1e3), max(self.bias*1e3), min(self.resistance), max(self.resistance)]

        self.axMap.set_ylabel("Resistance (Ohms)")
        self.axMap.set_xlabel("Bias (mV)")
        self.axMap.cla()
        self.draw()

    def cascade_plot(self,E_range,step,channel='didv'):
        plt.rcParams["axes.prop_cycle"] = plt.cycler("color", plt.cm.viridis(np.linspace(0,1,self.conductance.shape[0])))
        self.figure, self.ax = plt.subplots(1)
        
        count = 0
        if channel == 'didv':
            self.normalizeRange(E_range)
            for i in range(0,self.conductance.shape[0]):
                self.ax.plot(self.bias*1e3,self.conductance[i,:]+count*step)
                count += 1
        elif channel == 'current':
            for i in range(0,self.current.shape[0]):
                self.ax.plot(self.bias*1e3,self.current[i,:]/self.current[i,0]+count*step)
                count += 1
        plt.rcParams["axes.prop_cycle"] = plt.cycler("color", plt.cm.tab10(np.linspace(0,1,10)))


    def mapClick(self,event):
        if event.inaxes == self.axMap:
            self.energyCut = event.xdata
            self.cutPlotRange(self.energyCut)

    def update(self, val): #for the color scale sliders
        self.im1.set_clim([self.smin.val,self.smax.val])
        self.figure.canvas.draw()

    def cutPlotRange(self, energy):
        energy = energy*1e-3
        def LSinfluence_avg(id_c,id_n,id_p): #return contuctance averaged between different cuts
            idxs = np.arange(id_n,id_p)
            LSconductance_avg = np.zeros(len(self.conductance))
            if id_n == id_p: #checks that the range is >0
                LSconductance_avg = self.conductance[:,id_c]
            else:
                for idx in idxs:
                    LSconductance_avg = LSconductance_avg + self.conductance[:,idx]
                LSconductance_avg = LSconductance_avg/len(idxs)
            return LSconductance_avg
    
        #calculate the index based on the range given and the energy
        id_c = (abs(self.bias-energy)).argmin()
        id_n = (abs(self.bias-energy-self.rangecut)).argmin()
        id_p = (abs(self.bias-energy+self.rangecut)).argmin()
        print(id_c,id_n,id_p)
        self.conductance_avg = LSinfluence_avg(id_c,id_n,id_p)
        if id_n == id_p:
            self.axMap.plot([self.bias[id_c]*1e3,self.bias[id_c]*1e3],[self.resistance[0],self.resistance[-1]])
        else:
            self.axMap.fill_between([self.bias[id_n]*1e3,self.bias[id_p]*1e3],self.resistance[0],self.resistance[-1],alpha=0.6)
        self.axCut.plot(self.resistance,self.conductance_avg,label=str(round(self.bias[id_c]*1e3,3)))
        self.axCut.legend()
        self.saveCSV(self.resistance, self.conductance_avg)
        self.figure.canvas.draw_idle()

    def saveCSV(self, array1,array2):
        #for incremental save
        stridx = self.name[0].find('LS')
        LSidx = self.name[0][stridx:stridx+4]
        Eidx = np.round(self.energyCut,2)
        count = 0
        for i in os.listdir():
            if 'Cut{}'.format(count) in i:
                count += 1
        matrix = np.vstack((array1,array2))
        matrix = np.transpose(matrix)
        filename_mod = "Cut{}_{}_{}.txt".format(count,LSidx,Eidx)
        with open(filename_mod, 'w') as csvfile:
            writer = csv.writer(csvfile)
            [writer.writerow(r) for r in matrix]

    #todo:influence vertical cut, plot with tools


class grid():

    def __init__(self):
        pass
    def mapload(self,filename,cmap='viridis'):
        self.cmap = cmap
        self.gridraw = nanonis.grid()
        self.gridraw.load(filename)
    
    def explorer(self,mirror=False):
        self.mirror = mirror
        self.figure = plt.figure(figsize=(6,6))
        self.axMap = self.figure.add_subplot(1,1,1)
        self.figure.subplots_adjust(bottom=0.35)
        self.ax1 = self.figure.add_axes([0.20, 0.10, 0.65, 0.03])
        self.ax2 = self.figure.add_axes([0.20, 0.15, 0.65, 0.03])
        self.ax3 = self.figure.add_axes([0.20, 0.20, 0.65, 0.03])
        # self.energyCut_slider = Slider(self.ax1,'Energy cut',self.gridraw.bias.min()*1e3,self.gridraw.bias.max()*1e3,valinit=0, valstep=(self.gridraw.bias[0]-self.gridraw.bias[1])*1e3)
        self.energyCut_slider = Slider(self.ax1,'Energy cut',self.gridraw.bias.min()*1e3,self.gridraw.bias.max()*1e3,valinit=0, valstep=(0.01))
        if self.mirror == True:
            if 'LIX 1 omega (A)' in self.gridraw.data:
                self.conductance = np.flipud(self.gridraw.data['LIX 1 omega (A)'][:,:,0])
            if 'SRX (V)' in self.gridraw.data:
                self.conductance = np.flipud(self.gridraw.data['SRX (V)'][:,:,0])
            if 'LI Demod 1 X (A)' in self.gridraw.data:
                self.conductance = np.flipud(self.gridraw.data['LI Demod 1 X (A)'][:,:,0])
        else:
            if 'LIX 1 omega (A)' in self.gridraw.data:
                self.conductance = self.gridraw.data['LIX 1 omega (A)'][:,:,0]
            if 'LIX 1 omega (A)' in self.gridraw.data:
                self.conductance = self.gridraw.data['LIX 1 omega (A)'][:,:,0]
            if 'LI Demod 1 X (A)' in self.gridraw.data:
                self.conductance = self.gridraw.data['LI Demod 1 X (A)'][:,:,0]
        self.smin_slider = Slider(self.ax2, 'Min', self.conductance.min(), self.conductance.max(), valinit =self.conductance.min())
        self.smax_slider = Slider(self.ax3, 'Max', self.conductance.min(), self.conductance.max(), valinit =self.conductance.max()*0.5)            
        self.energyCut_slider.on_changed(self.update_energy)
        self.smin_slider.on_changed(self.update_cscale)
        self.smax_slider.on_changed(self.update_cscale)
        self.im1 = self.axMap.imshow(self.conductance,extent=[0,self.gridraw.xrange,0,self.gridraw.yrange],interpolation='nearest',cmap=self.cmap,vmax=0.5,vmin=0)
        
        #energy label 
        self.label = self.axMap.text(self.gridraw.xrange/10,self.gridraw.xrange/10,'0 mV',color='white')

        #axis labels
        self.axMap.set_xlabel('x (nm)')
        self.axMap.set_ylabel('y (nm)')


    def multicut(self,energies,vmins=None,vmaxs=None): #plot multiple grid cuts given the energies
        gsize = np.int(np.sqrt(len(energies)))
        grid = gs.GridSpec(gsize+1,gsize+1,wspace=0.3)
        self.figure = plt.figure(figsize = (10,10))
        count = 0
        for energy in energies:
            self.cutIdx = (abs(self.gridraw.bias-energy*1e-3)).argmin()
            self.conductance = np.flipud(self.gridraw.data['SRX (V)'][:,:,self.cutIdx])
            self.axMap = self.figure.add_subplot(grid[count])
            if vmaxs == None:
                self.im1 = self.axMap.imshow(self.conductance,extent=[0,self.gridraw.xrange,0,self.gridraw.yrange],interpolation=None,cmap=self.cmap,vmax=np.mean(self.conductance),vmin=self.conductance.min())
            else:
                self.im1 = self.axMap.imshow(self.conductance,extent=[0,self.gridraw.xrange,0,self.gridraw.yrange],interpolation=None,cmap=self.cmap,vmax=np.mean(self.conductance),vmin=self.conductance.min())
            count += 1

    def multicutFeTPP(self,energies,labels,vmins=None,vmaxs=None): #plot multiple grid cuts given the energies
        gsize = np.int(np.sqrt(len(energies)))
        grid = gs.GridSpec(gsize+1,gsize+1,wspace=0)
        self.figure = plt.figure(figsize = (10,10))
        count = 0
        for energy in energies:
            self.cutIdx = (abs(self.gridraw.bias-energy*1e-3)).argmin()
            self.conductance = np.flipud(self.gridraw.data['SRX (V)'][:-4,4:,self.cutIdx])
            self.axMap = self.figure.add_subplot(grid[count])
            if vmaxs == None:
                self.im1 = self.axMap.imshow(self.conductance,extent=[0,3,0,3],interpolation=None,cmap=self.cmap,vmin=0,vmax=None)
            else:
                self.im1 = self.axMap.imshow(self.conductance,extent=None,interpolation=None,cmap=self.cmap,vmin=0,vmax=vmaxs[count])

            self.axMap.get_xaxis().set_visible(False)
            self.axMap.get_yaxis().set_visible(False)
            #colorbar
            if count == 1 or count == 5 or count == 9:
                axin1 = self.axMap.inset_axes(self.axMap, width='100%', height='15%', loc='upper left',bbox_to_anchor=(-0.04,0.04,0.995,1), bbox_transform=self.axMap.transAxes)
            else:
                axin1 = self.axMap.inset_axes(self.axMap, width='100%', height='15%', loc='upper left',bbox_to_anchor=(-0.04,0.04,1,1), bbox_transform=self.axMap.transAxes)
                
            axin1.get_xaxis().set_visible(False)
            axin1.get_yaxis().set_visible(False)
            axin1.tick_params(axis='both',which='both',length=0)
            self.figure.colorbar(self.im1,cax=axin1,orientation='horizontal')
            #cbar label
            axin1.text(0.05,0.40,str(np.int(self.im1.get_clim()[0])),size=12,color='w',transform = axin1.transAxes,ha='center',va="center",weight='bold')
            axin1.text(0.90,0.40,str(np.round(self.im1.get_clim()[1],1)),size=12,color='k',transform = axin1.transAxes,ha='center',va="center",weight='bold')
            axin1.text(0.90,0.40,'',size=12,color='k',transform = axin1.transAxes,ha='center',va="center",weight='bold')
            # label text
            self.axMap.text(0,0.03,labels[count],color='w',size=14,weight='bold',transform = self.axMap.transAxes)
            self.axMap.text(0.03,0.03,str(np.round(self.gridraw.bias[self.cutIdx]*1e3,2))+'mV',color='w',size=12,weight='bold',transform = self.axMap.transAxes,ha='left')
            count += 1

    def update_energy(self,val):
        self.cutIdx = (abs(self.gridraw.bias-val*1e-3)).argmin()
        if self.mirror == True:
            if 'LIX 1 omega (A)' in self.gridraw.data:
                self.conductance = np.flipud(self.gridraw.data['LIX 1 omega (A)'][:,:,self.cutIdx])
            if 'SRX (V)' in self.gridraw.data:
                self.conductance = np.flipud(self.gridraw.data['SRX (V)'][:,:,self.cutIdx])
            if 'LI Demod 1 X (A)' in self.gridraw.data:
                self.conductance = np.flipud(self.gridraw.data['LI Demod 1 X (A)'][:,:,self.cutIdx])
        else:
            if 'LIX 1 omega (A)' in self.gridraw.data:
                self.conductance = self.gridraw.data['LIX 1 omega (A)'][:,:,self.cutIdx]
            if 'LIX 1 omega (A)' in self.gridraw.data:
                self.conductance = self.gridraw.data['LIX 1 omega (A)'][:,:,self.cutIdx]
            if 'LI Demod 1 X (A)' in self.gridraw.data:
                self.conductance = self.gridraw.data['LI Demod 1 X (A)'][:,:,self.cutIdx]
        self.im1.set_data(self.conductance)
        self.im1.set_clim(np.min(self.conductance),np.max(self.conductance))
        self.label.set_text('{} mV'.format(np.round(val,2)))
        self.figure.canvas.draw()

    def update_cscale(self,val):
        self.im1.set_clim([self.smin_slider.val,self.smax_slider.val])
        self.figure.canvas.draw()


class sliderImage():
    def __init__(self) -> None:
        pass

    def load(self,map):
        self.map = map

    def plot(self,slim=None,valinit=None):
        self.figure = plt.figure(figsize = (5,5))
        self.figure.subplots_adjust(bottom=0.3)
        self.axMap = self.figure.add_subplot(111)
        self.axmin = self.figure.add_axes([0.15, 0.1, 0.65, 0.03])
        self.axmax = self.figure.add_axes([0.15, 0.15, 0.65, 0.03])

        if slim == None:
            slim = (np.min(self.map)/2,np.max(self.map)*2)
            valinit = np.max(self.map)
        self.smin = Slider(self.axmin, 'Min', slim[0] ,slim[1] , valinit =valinit)
        self.smax = Slider(self.axmax, 'Max', slim[0] ,slim[1] , valinit =valinit)
        self.smin.on_changed(self.update)
        self.smax.on_changed(self.update)

        self.im1 = self.axMap.imshow(self.map,vmin=slim[0],vmax=slim[1],aspect='auto',interpolation='nearest',cmap='Blues')



    def update(self, val): #for the color scale sliders
        self.im1.set_clim([self.smin.val,self.smax.val])
        self.figure.canvas.draw()

class ZapproachMilano():

    def __init__(self):
        pass
    def load(self,filenames):
        self.Zapproach = nanonis.Zapproach()
        self.Zapproach.load(filenames)
    
    def mapPlot(self):
        fig,ax = plt.subplots(1)
        ax.imshow(self.Zapproach.conductance,extent=[self.Zapproach.bias[-1],self.Zapproach.bias[0],0,1],aspect='auto',interpolation='nearest')
        ax.set_xlabel('Bias (mV)')

##useful plotting tools

#set the current axes in spectroscopy mode

class tri_grid():
#Falta hacer bien el MLS
    def __init__(self):
        self.type = 'Grid'
    

    def load(self,fnames,G0_set_fname,normalizeR=[4e-3,5e-3]):
        # create triangle points and load centers
        self.G0_set = uf.load_obj(G0_set_fname)
        v1 = self.G0_set['v1']
        v2 = self.G0_set['v2']
        v3 = self.G0_set['v3']
        center_x = self.G0_set['center_x'] 
        center_y = self.G0_set['center_y'] 
        width = self.G0_set['width']
        self.width = width
        height = self.G0_set['height']
        self.height = height
        spacing = self.G0_set['spacing']
        square_points = self.generate_square_grid(center_x, center_y, width, height, spacing)
        self.triangle_points = self.cut_triangle_from_square(square_points, v1, v2, v3)
        self.triangle_points = self.sort_points(self.triangle_points)
        self.x = np.arange(center_x-width/2,center_x+width/2,spacing)
        self.y = np.arange(center_y-width/2,center_y+width/2,spacing)
        n=0
        self.spectra = nanonis.biasSpectroscopy()
        self.spectra.load(fnames[0])
        self.bias = self.spectra.bias
        self.tri_grid = np.zeros((self.x.shape[0],self.y.shape[0],self.bias.shape[0]))
        for f in fnames[:-1]:
            x_idx = np.abs(self.x-self.triangle_points[n][0]).argmin()
            y_idx = np.abs(self.y-self.triangle_points[n][1]).argmin()
            self.spectra.load(f)
            # self.spectra.normalizeRange(normalizeR)
            self.tri_grid[x_idx,y_idx,:] = self.spectra.conductance
            print(x_idx,y_idx)
            n+=1

        self.x_coords = np.linspace(0,self.width,self.tri_grid.shape[0])
        self.y_coords = np.linspace(0,self.height,self.tri_grid.shape[1])
        return



    def explorer(self):
        self.figure = plt.figure(figsize=(6,6))
        self.axMap = self.figure.add_subplot(211)
        self.axSpec = self.figure.add_subplot(212)
        self.axSpec.margins(0.05)
        self.axMap.margins(0.05)
        self.figure.subplots_adjust(bottom=0.35)
        self.ax1 = self.figure.add_axes([0.20, 0.10, 0.65, 0.03])
        self.ax2 = self.figure.add_axes([0.20, 0.15, 0.65, 0.03])
        self.ax3 = self.figure.add_axes([0.20, 0.20, 0.65, 0.03])
        self.conductance = self.tri_grid[:,:,0]
        # self.energyCut_slider = Slider(self.ax1,'Energy cut',self.gridraw.bias.min()*1e3,self.gridraw.bias.max()*1e3,valinit=0, valstep=(self.gridraw.bias[0]-self.gridraw.bias[1])*1e3)
        self.energyCut_slider = Slider(self.ax1,'Energy cut',self.bias.min()*1e3,self.bias.max()*1e3,valinit=0, valstep=(0.01))
        self.smin_slider = Slider(self.ax2, 'Min', self.conductance.min(), self.conductance.max(), valinit =self.conductance.min())
        self.smax_slider = Slider(self.ax3, 'Max', self.conductance.min(), self.conductance.max(), valinit =self.conductance.max())            
        self.energyCut_slider.on_changed(self.update_energy)
        self.smin_slider.on_changed(self.update_cscale)
        self.smax_slider.on_changed(self.update_cscale)
        self.im1 = self.axMap.imshow(np.rot90(np.rot90(np.rot90(self.conductance))),extent=[0,self.width,0,self.height],interpolation='nearest',cmap='viridis',vmax=0.5,vmin=0)
        self.axvline = self.axSpec.axvline(0)
        #energy label
        self.label = self.axMap.text(self.width/10,self.height/10,'0 mV',color='white')
        #axis labels
        self.axMap.set_xlabel('x (nm)')
        self.axMap.set_ylabel('y (nm)')
        # plot of the spectra
        self.axSpec.plot(self.bias*1e3,np.mean(self.tri_grid,axis=(0,1)))
        self.figure.canvas.mpl_connect('button_press_event', self.update_spectrum_on_click)

    def update_energy(self,val):
        self.cutIdx = (abs(self.bias-val*1e-3)).argmin()
        self.conductance = self.tri_grid[:,:,self.cutIdx]
        self.im1.set_data(np.rot90(self.conductance))
        self.im1.set_clim(np.min(self.conductance),np.max(self.conductance))
        self.label.set_text('{} mV'.format(np.round(val,2)))
        self.figure.canvas.draw()
        # put index in plot
        self.axvline.remove()
        self.axvline = self.axSpec.axvline(val)

        
    def update_cscale(self,val):
        self.im1.set_clim([self.smin_slider.val,self.smax_slider.val])
        self.figure.canvas.draw()

    def update_spectrum_on_click(self, event):
        if event.inaxes == self.axMap:
            x_coord, y_coord = event.xdata, event.ydata
            x_idx = np.abs(self.x_coords - y_coord).argmin()
            y_idx = np.abs(self.y_coords - x_coord).argmin()
            selected_spectrum = self.tri_grid[x_idx,y_idx ,:]
            
            # Update the spectra plot
            self.axSpec.clear()
            self.axSpec.plot(self.bias * 1e3, selected_spectrum)
            self.axSpec.set_xlabel('Bias (mV)')
            self.axSpec.set_ylabel('Conductance')
            self.axSpec.set_title(f'Spectrum at ({x_coord}, {y_coord})')
            self.figure.canvas.draw()


    def frange(self,start, end, step):
        while start <= end:
            yield start
            start += step

    def generate_square_grid(self,center_x, center_y, width, height, spacing=1.0):
        min_x = center_x - width / 2
        max_x = center_x + width / 2
        min_y = center_y - height / 2
        max_y = center_y + height / 2
        square_points = []
        for x in self.frange(min_x, max_x, spacing):
            for y in self.frange(min_y, max_y, spacing):
                square_points.append((x, y))

        return square_points

    def is_inside_triangle(self,point, v1, v2, v3):
        def sign(p1, p2, p3):
            return (p1[0] - p3[0]) * (p2[1] - p3[1]) - (p2[0] - p3[0]) * (p1[1] - p3[1])
        b1 = sign(point, v1, v2) < 0.0
        b2 = sign(point, v2, v3) < 0.0
        b3 = sign(point, v3, v1) < 0.0

        return b1 == b2 == b3

    def cut_triangle_from_square(self,square_points, v1, v2, v3):
        triangle_points = []
        for point in square_points:
            if self.is_inside_triangle(point, v1, v2, v3):
                triangle_points.append(point)
        return triangle_points

    def sort_points(self,points):
        # Sort points first by -X (right to left), then by Y (bottom to top)
        return sorted(points, key=lambda point: (point[1], -point[0]))


class tri_grid_new():
#Falta hacer bien el MLS
    def __init__(self):
        self.type = 'Grid'
    
    def load(self,fnames,triangle_points,square_points,center_x,center_y,width,height,normalizeR=[4e-3,5e-3]):
        x = np.arange(center_x-width/2,center_x+width/2,1.5)
        y = np.arange(center_y-width/2,center_y+width/2,1.5)
        self.tri_grid = np.zeros((x.shape[0],y.shape[0],300))+0
        self.x = width
        self.y = height
        n=0
        self.spectra = nanonis.biasSpectroscopy()
        self.spectra.load(fnames[0])
        self.bias = self.spectra.bias
        for f in fnames[:-1]:
            x_idx = np.abs(x-triangle_points[n][0]).argmin()
            y_idx = np.abs(y-triangle_points[n][1]).argmin()
            self.spectra.load(f)
            self.spectra.normalizeRange(normalizeR)
            self.tri_grid[x_idx,y_idx,:] = self.spectra.conductance
            n+=1
        return
    def explorer(self):
        self.figure = plt.figure(figsize=(6,6))
        self.axMap = self.figure.add_subplot(211)
        self.axSpec = self.figure.add_subplot(212)
        self.axSpec.margins(0.05)
        self.axMap.margins(0.05)
        self.figure.subplots_adjust(bottom=0.35)
        self.ax1 = self.figure.add_axes([0.20, 0.10, 0.65, 0.03])
        self.ax2 = self.figure.add_axes([0.20, 0.15, 0.65, 0.03])
        self.ax3 = self.figure.add_axes([0.20, 0.20, 0.65, 0.03])
        self.conductance = self.tri_grid[:,:,0]
        # self.energyCut_slider = Slider(self.ax1,'Energy cut',self.gridraw.bias.min()*1e3,self.gridraw.bias.max()*1e3,valinit=0, valstep=(self.gridraw.bias[0]-self.gridraw.bias[1])*1e3)
        self.energyCut_slider = Slider(self.ax1,'Energy cut',self.bias.min()*1e3,self.bias.max()*1e3,valinit=0, valstep=(0.01))
        self.smin_slider = Slider(self.ax2, 'Min', self.conductance.min(), self.conductance.max(), valinit =self.conductance.min())
        self.smax_slider = Slider(self.ax3, 'Max', self.conductance.min(), self.conductance.max(), valinit =self.conductance.max()*0.5)            
        self.energyCut_slider.on_changed(self.update_energy)
        self.smin_slider.on_changed(self.update_cscale)
        self.smax_slider.on_changed(self.update_cscale)
        self.im1 = self.axMap.imshow(self.conductance,extent=[0,self.x,0,self.y],interpolation='nearest',cmap='viridis',vmax=0.5,vmin=0)
        #energy label 
        self.label = self.axMap.text(self.x/10,self.x/10,'0 mV',color='white')
        #axis labels
        self.axMap.set_xlabel('x (nm)')
        self.axMap.set_ylabel('y (nm)')
        # plot of the spectra
        self.axSpec.plot(self.bias*1e3,np.mean(self.tri_grid,axis=(0,1)))


    def update_energy(self,val):
        self.cutIdx = (abs(self.bias-val*1e-3)).argmin()
        self.conductance = np.flipud(self.tri_grid[:,:,self.cutIdx])
        self.im1.set_data(self.conductance)
        self.im1.set_clim(np.min(self.conductance),np.max(self.conductance))
        self.label.set_text('{} mV'.format(np.round(val,2)))
        self.figure.canvas.draw()
        # put index in plot
        self.axSpec.clear()
        self.axSpec.plot(self.bias*1e3,np.mean(self.tri_grid,axis=(0,1)))
        self.axSpec.axvline(val)
    def update_cscale(self,val):
        self.im1.set_clim([self.smin_slider.val,self.smax_slider.val])
        self.figure.canvas.draw()



class tri_grid_fix():
#Falta hacer bien el MLS
    def __init__(self):
        self.type = 'Grid'
    

    def load(self,fnames,square_points):

        self.triangle_points = self.cut_triangle_from_square(square_points, v1, v2, v3)
        self.triangle_points = self.sort_points(self.triangle_points)
        self.x = np.arange(center_x-width/2,center_x+width/2,spacing)
        self.y = np.arange(center_y-width/2,center_y+width/2,spacing)
        n=0
        self.spectra = nanonis.biasSpectroscopy()
        self.spectra.load(fnames[0])
        self.bias = self.spectra.bias
        self.tri_grid = np.zeros((self.x.shape[0],self.y.shape[0],self.bias.shape[0]))
        for f in fnames[:-1]:
            x_idx = np.abs(self.x-self.triangle_points[n][0]).argmin()
            y_idx = np.abs(self.y-self.triangle_points[n][1]).argmin()
            self.spectra.load(f)
            # self.spectra.normalizeRange(normalizeR)
            self.tri_grid[x_idx,y_idx,:] = self.spectra.conductance
            print(x_idx,y_idx)
            n+=1

        self.x_coords = np.linspace(0,self.width,self.tri_grid.shape[0])
        self.y_coords = np.linspace(0,self.height,self.tri_grid.shape[1])
        return



    def explorer(self):
        self.figure = plt.figure(figsize=(6,6))
        self.axMap = self.figure.add_subplot(211)
        self.axSpec = self.figure.add_subplot(212)
        self.axSpec.margins(0.05)
        self.axMap.margins(0.05)
        self.figure.subplots_adjust(bottom=0.35)
        self.ax1 = self.figure.add_axes([0.20, 0.10, 0.65, 0.03])
        self.ax2 = self.figure.add_axes([0.20, 0.15, 0.65, 0.03])
        self.ax3 = self.figure.add_axes([0.20, 0.20, 0.65, 0.03])
        self.conductance = self.tri_grid[:,:,0]
        # self.energyCut_slider = Slider(self.ax1,'Energy cut',self.gridraw.bias.min()*1e3,self.gridraw.bias.max()*1e3,valinit=0, valstep=(self.gridraw.bias[0]-self.gridraw.bias[1])*1e3)
        self.energyCut_slider = Slider(self.ax1,'Energy cut',self.bias.min()*1e3,self.bias.max()*1e3,valinit=0, valstep=(0.01))
        self.smin_slider = Slider(self.ax2, 'Min', self.conductance.min(), self.conductance.max(), valinit =self.conductance.min())
        self.smax_slider = Slider(self.ax3, 'Max', self.conductance.min(), self.conductance.max(), valinit =self.conductance.max())            
        self.energyCut_slider.on_changed(self.update_energy)
        self.smin_slider.on_changed(self.update_cscale)
        self.smax_slider.on_changed(self.update_cscale)
        self.im1 = self.axMap.imshow(np.rot90(np.rot90(np.rot90(self.conductance))),extent=[0,self.width,0,self.height],interpolation='nearest',cmap='viridis',vmax=0.5,vmin=0)
        self.axvline = self.axSpec.axvline(0)
        #energy label
        self.label = self.axMap.text(self.width/10,self.height/10,'0 mV',color='white')
        #axis labels
        self.axMap.set_xlabel('x (nm)')
        self.axMap.set_ylabel('y (nm)')
        # plot of the spectra
        self.axSpec.plot(self.bias*1e3,np.mean(self.tri_grid,axis=(0,1)))
        self.figure.canvas.mpl_connect('button_press_event', self.update_spectrum_on_click)

    def update_energy(self,val):
        self.cutIdx = (abs(self.bias-val*1e-3)).argmin()
        self.conductance = self.tri_grid[:,:,self.cutIdx]
        self.im1.set_data(np.rot90(self.conductance))
        self.im1.set_clim(np.min(self.conductance),np.max(self.conductance))
        self.label.set_text('{} mV'.format(np.round(val,2)))
        self.figure.canvas.draw()
        # put index in plot
        self.axvline.remove()
        self.axvline = self.axSpec.axvline(val)

        
    def update_cscale(self,val):
        self.im1.set_clim([self.smin_slider.val,self.smax_slider.val])
        self.figure.canvas.draw()

    def update_spectrum_on_click(self, event):
        if event.inaxes == self.axMap:
            x_coord, y_coord = event.xdata, event.ydata
            x_idx = np.abs(self.x_coords - y_coord).argmin()
            y_idx = np.abs(self.y_coords - x_coord).argmin()
            selected_spectrum = self.tri_grid[x_idx,y_idx ,:]
            
            # Update the spectra plot
            self.axSpec.clear()
            self.axSpec.plot(self.bias * 1e3, selected_spectrum)
            self.axSpec.set_xlabel('Bias (mV)')
            self.axSpec.set_ylabel('Conductance')
            self.axSpec.set_title(f'Spectrum at ({x_coord}, {y_coord})')
            self.figure.canvas.draw()


    def frange(self,start, end, step):
        while start <= end:
            yield start
            start += step

    def generate_square_grid(self,center_x, center_y, width, height, spacing=1.0):
        min_x = center_x - width / 2
        max_x = center_x + width / 2
        min_y = center_y - height / 2
        max_y = center_y + height / 2
        square_points = []
        for x in self.frange(min_x, max_x, spacing):
            for y in self.frange(min_y, max_y, spacing):
                square_points.append((x, y))

        return square_points

    def is_inside_triangle(self,point, v1, v2, v3):
        def sign(p1, p2, p3):
            return (p1[0] - p3[0]) * (p2[1] - p3[1]) - (p2[0] - p3[0]) * (p1[1] - p3[1])
        b1 = sign(point, v1, v2) < 0.0
        b2 = sign(point, v2, v3) < 0.0
        b3 = sign(point, v3, v1) < 0.0

        return b1 == b2 == b3

    def cut_triangle_from_square(self,square_points, v1, v2, v3):
        triangle_points = []
        for point in square_points:
            if self.is_inside_triangle(point, v1, v2, v3):
                triangle_points.append(point)
        return triangle_points

    def sort_points(self,points):
        # Sort points first by -X (right to left), then by Y (bottom to top)
        return sorted(points, key=lambda point: (point[1], -point[0]))



import numpy as np
from scipy import constants as const


Eh = const.physical_constants['Hartree energy in eV']
## Improoved No PoF Superconductivity
def initD_(V,delta_t,Delta,T,eta,eta_t):
    global EnD
    EnD=[0 for i in range(25)]
    global MnD
    MnD=[0 for i in range(25)]
    global x_int
    x_int=np.linspace(-4*np.max(V),4*np.max(V),5000)

    global bcsf
    bcsf=bcs(delta_t,x_int,eta_t)*fermi(T,x_int)

    global bcsif_even
    bcsif_even=bcs_i(Delta,x_int,eta,0)*fermi(T,x_int)

    global bcsif_odd
    bcsif_odd=bcs_i(Delta,x_int,eta,1)*fermi(T,x_int)

def bcs(delta,x,eta):
    if delta<=0.01: 
        return np.full(np.shape(x),1)
    else:
        return (np.sign(x))*np.imag(np.divide(np.abs(x+eta*1j),np.sqrt(delta**2-(x+eta*1j)**2)))
    
def bcs_i(delta,x,eta,n):
    if np.mod(n,2)==0:
        if delta<=0.01: 
            return np.full(np.shape(x),1)
        else:
            return (np.sign(x))*np.imag(np.divide(np.abs(x+eta*1j),np.sqrt(delta**2-(x+eta*1j)**2)))
    else:
        return 1
        

def fermi(T,x):
    if T==0.0:
        return np.heaviside(-x,1)
    else:
        return np.divide(1,1+np.exp(x/T))

def E0(C1,C2,V,n,Q0,Delta):
    return (n+Q0)**2/(2*(C1+C2))+(1-(-1)**(n))*Delta/2

def E1(C1,C2,V,n,Q0):
    k=C2/(C1+C2)
    return k*V+(n+Q0-1/2)/(C1+C2)

def E2(C1,C2,V,n,Q0):
    k=C1/(C1+C2)
    return k*V+(n+Q0-1/2)/(C1+C2)


def Gamma1D(V,R1,C1,C2,n,Q0,Delta,delta_t,delta_s,T,eta):
    a,b=np.meshgrid(x_int,E1(C1,C2,V,n,Q0))
    t=a+b
    return np.dot( bcs_i(Delta,t,eta,n)*(fermi(T,-t)),bcsf )/R1

def Gamma2D(V,R2,C1,C2,n,Q0,delta_s,delta_t,T,eta):
    a,b=np.meshgrid(x_int,E2(C1,C2,V,-n,-Q0))
    t=a+b
    if np.mod(n,2)==0:
        return np.dot(bcs(delta_s,t,eta)*(fermi(T,-t)),bcsif_even)/R2
    else:
        return np.dot(bcs(delta_s,t,eta)*(fermi(T,-t)),bcsif_odd)/R2



def PND(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta,n):
    mn=(Gamma1D(V,R1,C1,C2,-n,Q0,Delta,delta_t,delta_s,T,eta)+Gamma2D(-V,R2,C1,C2,n,-Q0,Delta,delta_t,delta_s,T,eta))/(Gamma1D(-V,R1,C1,C2,n+1,-Q0,Delta,delta_t,delta_s,T,eta)+Gamma2D(V,R2,C1,C2,-n-1,Q0,Delta,delta_t,delta_s,T,eta))
    en=(Gamma1D(-V,R1,C1,C2,-n,-Q0,Delta,delta_t,delta_s,T,eta)+Gamma2D(V,R2,C1,C2,n,Q0,Delta,delta_t,delta_s,T,eta))/(Gamma1D(V,R1,C1,C2,n+1,Q0,Delta,delta_t,delta_s,T,eta)+Gamma2D(-V,R2,C1,C2,-n-1,-Q0,Delta,delta_t,delta_s,T,eta))
    return en,mn

def check_pD(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta):
    V=V[-1]
    n=[0]
    while True:
        Z=0
        for i in range(-n[-1],n[-1]+1):
            Z+=np.exp(-E0(C1,C2,V,i,Q0,Delta)/T)
        pn=np.exp(-E0(C1,C2,V,n[-1],Q0,Delta)/T)/Z
        p_n=np.exp(-E0(C1,C2,V,-n[-1],Q0,Delta)/T)/Z
        n.append(n[-1]+1)
        if pn<0.01 and p_n<0.01:
            #print(len(a))
            break 
    return n[-1]

def all_pD(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta,n):
    Z=0
    for i in range(-n,n+1):
        Z+=np.exp(-E0(C1,C2,V,i,Q0,Delta)/T)
    p0=np.exp(-E0(C1,C2,V,0,Q0,Delta)/T)/Z
    pn=[]
    p_n=[]
    for i in range(1,n+1):
        pn.append(np.exp(-E0(C1,C2,V,i,Q0,Delta)/T)/Z)
        p_n.append(np.exp(-E0(C1,C2,V,-i,Q0,Delta)/T)/Z)
    return p0,pn,p_n

def G1nD(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta,n):
    return Gamma1D(V,R1,C1,C2,n,Q0,Delta,delta_t,delta_s,T,eta)-Gamma1D(-V,R1,C1,C2,-n,-Q0,Delta,delta_t,delta_s,T,eta)

def G2nD(V,R1,R2,C1,C2,Q0,Delta,delta_t,delta_s,T,eta,n):
    return Gamma2D(V,R2,C1,C2,n,Q0,Delta,delta_t,delta_s,T,eta)-Gamma2D(-V,R2,C1,C2,-n,-Q0,Delta,delta_t,delta_s,T,eta)

def currentD(V,R1,R2,C1,C2,Q0,Delta,Delta_t,Delta_s,T,eta,eta_t):
    initD_(V,Delta_t,Delta,T,eta,eta_t)
    n=check_pD(V,R1,R2,C1,C2,Q0,Delta,Delta_t,Delta_s,T,eta)
    p0,pn,p_n=all_pD(V,R1,R2,C1,C2,Q0,Delta,Delta_t,Delta_s,T,eta,n)
    I=p0*G1nD(V,R1,R2,C1,C2,Q0,Delta,Delta_t,Delta_s,T,eta,0)
    for i in range(1,n+1):
        I+=pn[i-1]*G1nD(V,R1,R2,C1,C2,Q0,Delta,Delta_t,Delta_s,T,eta,i)
        I+=p_n[i-1]*G1nD(V,R1,R2,C1,C2,Q0,Delta,Delta_t,Delta_s,T,eta,-i)
    t=np.gradient(I)
    return t/t[0]


def currentD2(V,R1,R2,C1,C2,Q0,Delta,Delta_t,Delta_s,T,eta,eta_t):
    initD_(V,Delta_t,Delta,T,eta,eta_t)
    n=check_pD(V,R1,R2,C1,C2,Q0,Delta,Delta_t,Delta_s,T,eta)
    p0,pn,p_n=all_pD(V,R1,R2,C1,C2,Q0,Delta,Delta_t,Delta_s,T,eta,n)
    I=p0*G2nD(V,R1,R2,C1,C2,Q0,Delta,Delta_t,Delta_s,T,eta,0)
    for i in range(1,n+1):
        I+=pn[i-1]*G2nD(V,R1,R2,C1,C2,Q0,Delta,Delta_t,Delta_s,T,eta,i)
        I+=p_n[i-1]*G2nD(V,R1,R2,C1,C2,Q0,Delta,Delta_t,Delta_s,T,eta,-i)
    t=np.gradient(I)
    return t/t[0]


import matplotlib.pyplot as plt
import nanonis

spectra = nanonis.biasSpectroscopy()

def fitter(fname,initial_params=None,off=0):
    fnames = fname
    if initial_params == None:
            p = {'R1': 1000,
        'R2': 1,
        'C1': 0.002,
        'C2': 3,
        'T': 0.07,
        'Q0': -0.75,
        'Delta': 1.35,
        'Delta_t': 1.35,
        'Delta_s': 0.3,
        'eta': 0.08,
        'eta_t': 0.08}
    else:
        p = initial_params


    plt.rcParams["axes.prop_cycle"] = plt.cycler("color", plt.cm.tab10(np.linspace(0,1,10)))

    last=True




    fig,ax=plt.subplots()
    fig.subplots_adjust(bottom=0.7)

    def plotdata(fnames):
        for f in fnames:
            spectra.load(f)
            spectra.biasOffset(off*1e-3)
            spectra.normalizeRange_symm([4e-3,5e-3])
            b1, c1 = spectra.bias*1e3,spectra.conductance
            ax.plot(b1, c1)

    plotdata(fnames)

    # current calc
    x=np.linspace(-5,5,100)
    y=currentD(x,300,1,0.01,0.2,0.0,1.3,1.3,0.1,0.1,0.05,0.05)
    ax.plot(x,y/y[0])

    ax1 = fig.add_axes([0.10, 0.10, 0.65, 0.03])
    ax2 = fig.add_axes([0.10, 0.15, 0.65, 0.03])
    ax3 = fig.add_axes([0.10, 0.20, 0.65, 0.03])
    ax4 = fig.add_axes([0.10, 0.25, 0.65, 0.03])
    ax5 = fig.add_axes([0.10, 0.30, 0.65, 0.03])
    ax6 = fig.add_axes([0.10, 0.35, 0.65, 0.03])
    ax7 = fig.add_axes([0.10, 0.40, 0.65, 0.03])
    ax8 = fig.add_axes([0.10, 0.45, 0.65, 0.03])
    ax9 = fig.add_axes([0.10, 0.5, 0.65, 0.03])
    ax10 = fig.add_axes([0.10,0.55, 0.65, 0.03])
    ax11 = fig.add_axes([0.10,0.60, 0.65, 0.03])


    if last==False:
        ax1_s = plt.Slider(ax1,'R1',1,1000,valinit=300)
        ax2_s = plt.Slider(ax2,'R2',1,1000,valinit=1)
        ax3_s = plt.Slider(ax3,'C1',0.00001,0.2,valinit=0.01)
        ax4_s = plt.Slider(ax4,'C2',0.001,5,valinit=0.1)
        ax5_s = plt.Slider(ax5,'T',0,1,valinit=0.1)
        ax6_s = plt.Slider(ax6,'Q0',-1,1,valinit=0)
        ax7_s = plt.Slider(ax7,'Delta',0,1.3,valinit=1.3)
        ax8_s = plt.Slider(ax8,'Delta_t',0,1.5,valinit=1.3)
        ax9_s = plt.Slider(ax9,'Delta_s',0,1.3,valinit=0.3)
        ax10_s = plt.Slider(ax10,'eta',0.01,0.2,valinit=0.05)
        ax11_s = plt.Slider(ax11,'eta_t',0.01,0.2,valinit=0.05)



    if last==True:
        params = p
        # params = {
        # 'R1' : ax1_s.val,
        # 'R2' : ax2_s.val,
        # 'C1' : ax3_s.val,
        # 'C2' : ax4_s.val,
        # 'T' : ax5_s.val,
        # 'Q0' : ax6_s.val,
        # 'Delta' : ax7_s.val,
        # 'Delta_t' : ax8_s.val,
        # 'Delta_s' : ax9_s.val,
        # 'eta' : ax10_s.val
        # }
        ax1_s = plt.Slider(ax1,'R1',1,1000,valinit=params['R1'])
        ax2_s = plt.Slider(ax2,'R2',1,1000,valinit=params['R2'])
        ax3_s = plt.Slider(ax3,'C1',0.000001,0.2,valinit=params['C1'])
        ax4_s = plt.Slider(ax4,'C2',0.001,5,valinit=params['C2'])
        ax5_s = plt.Slider(ax5,'T',0,1,valinit=params['T'])
        ax6_s = plt.Slider(ax6,'Q0',-2,2,valinit=params['Q0'])
        ax7_s = plt.Slider(ax7,'Delta',0,1.5,valinit=params['Delta'])
        ax8_s = plt.Slider(ax8,'Delta_t',0,1.5,valinit=params['Delta_t'])
        ax9_s = plt.Slider(ax9,'Delta_s',0,1.5,valinit=params['Delta_s'])
        ax10_s = plt.Slider(ax10,'eta',0.01,0.2,valinit=params['eta'])
        ax11_s = plt.Slider(ax11,'eta_t',0.01,0.2,valinit=0.05)

    def update(val):
        ax.clear()
        y=currentD(x,ax1_s.val,ax2_s.val,ax3_s.val,ax4_s.val,ax6_s.val,ax7_s.val,ax8_s.val,ax9_s.val,ax5_s.val,ax10_s.val,ax11_s.val)
        ax.plot(x,y/y[0],color='k')
        plotdata(fnames)
        params = {
        'R1' : ax1_s.val,
        'R2' : ax2_s.val,
        'C1' : ax3_s.val,
        'C2' : ax4_s.val,
        'T' : ax5_s.val,
        'Q0' : ax6_s.val,
        'Delta' : ax7_s.val,
        'Delta_t' : ax8_s.val,
        'Delta_s' : ax9_s.val,
        'eta' : ax10_s.val
        }

    ax1_s.on_changed(update)
    ax2_s.on_changed(update)
    ax3_s.on_changed(update)
    ax4_s.on_changed(update)
    ax5_s.on_changed(update)
    ax6_s.on_changed(update)
    ax7_s.on_changed(update)
    ax8_s.on_changed(update)
    ax9_s.on_changed(update)
    ax10_s.on_changed(update)
    ax11_s.on_changed(update)

    return


def E_odd(V,n,C,Delta):
    return (n-V*C)**2/(2*C)+Delta

def E_even(V,n,C):
    return (n-V*C)**2/(2*C)
